from __future__ import annotations

from inspect import Parameter, Signature
from typing import List, Union  # NoQA: UP035


class Foo:
    pass


class Bar:
    def __init__(self, x, y):
        pass


class Baz:
    def __new__(cls, x, y):
        pass


class Qux:
    __signature__ = Signature(
        parameters=[
            Parameter('foo', Parameter.POSITIONAL_OR_KEYWORD),
            Parameter('bar', Parameter.POSITIONAL_OR_KEYWORD),
        ]
    )

    def __init__(self, x, y):
        pass


class Quux(List[Union[int, float]]):  # NoQA: UP006,UP007
    """A subclass of List[Union[int, float]]"""

    pass


class Corge(Quux):
    pass


Alias = Foo

#: docstring
OtherAlias = Bar

#: docstring
IntAlias = int



from __future__ import annotations

import zlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Final

INVENTORY_V1: Final[bytes] = b"""\
# Sphinx inventory version 1
# Project: foo
# Version: 1.0
module mod foo.html
module.cls class foo.html
"""

INVENTORY_V2: Final[bytes] = b"""\
# Sphinx inventory version 2
# Project: foo
# Version: 2.0
# The remainder of this file is compressed with zlib.
""" + zlib.compress(b"""\
module1 py:module 0 foo.html#module-module1 Long Module desc
module2 py:module 0 foo.html#module-$ -
module1.func py:function 1 sub/foo.html#$ -
module1.Foo.bar py:method 1 index.html#foo.Bar.baz -
CFunc c:function 2 cfunc.html#CFunc -
std cpp:type 1 index.html#std -
std::uint8_t cpp:type 1 index.html#std_uint8_t -
foo::Bar cpp:class 1 index.html#cpp_foo_bar -
foo::Bar::baz cpp:function 1 index.html#cpp_foo_bar_baz -
foons cpp:type 1 index.html#foons -
foons::bartype cpp:type 1 index.html#foons_bartype -
a term std:term -1 glossary.html#term-a-term -
ls.-l std:cmdoption 1 index.html#cmdoption-ls-l -
docname std:doc -1 docname.html -
foo js:module 1 index.html#foo -
foo.bar js:class 1 index.html#foo.bar -
foo.bar.baz js:method 1 index.html#foo.bar.baz -
foo.bar.qux js:data 1 index.html#foo.bar.qux -
a term including:colon std:term -1 glossary.html#term-a-term-including-colon -
The-Julia-Domain std:label -1 write_inventory/#$ The Julia Domain
""")

INVENTORY_V2_NO_VERSION: Final[bytes] = b"""\
# Sphinx inventory version 2
# Project: foo
# Version:
# The remainder of this file is compressed with zlib.
""" + zlib.compress(b"""\
module1 py:module 0 foo.html#module-module1 Long Module desc
""")

INVENTORY_V2_AMBIGUOUS_TERMS: Final[bytes] = b"""\
# Sphinx inventory version 2
# Project: foo
# Version: 2.0
# The remainder of this file is compressed with zlib.
""" + zlib.compress(b"""\
a term std:term -1 glossary.html#term-a-term -
A term std:term -1 glossary.html#term-a-term -
b term std:term -1 document.html#id5 -
B term std:term -1 document.html#B -
""")



from __future__ import annotations

import hashlib
import re
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='root')
def test_html_download(app: SphinxTestApp) -> None:
    app.build()

    # subdir/includes.html
    result = (app.outdir / 'subdir' / 'includes.html').read_text(encoding='utf8')
    pattern = (
        '<a class="reference download internal" download="" '
        'href="../(_downloads/.*/img.png)">'
    )
    matched = re.search(pattern, result)
    assert matched
    assert (app.outdir / matched.group(1)).exists()
    filename = matched.group(1)

    # includes.html
    result = (app.outdir / 'includes.html').read_text(encoding='utf8')
    pattern = (
        '<a class="reference download internal" download="" '
        'href="(_downloads/.*/img.png)">'
    )
    matched = re.search(pattern, result)
    assert matched
    assert (app.outdir / matched.group(1)).exists()
    assert matched.group(1) == filename

    pattern = (
        '<a class="reference download internal" download="" '
        'href="(_downloads/.*/)(file_with_special_%23_chars.xyz)">'
    )
    matched = re.search(pattern, result)
    assert matched
    assert (app.outdir / matched.group(1) / 'file_with_special_#_chars.xyz').exists()


@pytest.mark.sphinx('html', testroot='roles-download')
def test_html_download_role(app: SphinxTestApp) -> None:
    app.build()
    digest = hashlib.md5(b'dummy.dat', usedforsecurity=False).hexdigest()
    assert (app.outdir / '_downloads' / digest / 'dummy.dat').exists()
    digest_another = hashlib.md5(
        b'another/dummy.dat', usedforsecurity=False
    ).hexdigest()
    assert (app.outdir / '_downloads' / digest_another / 'dummy.dat').exists()

    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<li><p><a class="reference download internal" download="" '
        'href="_downloads/%s/dummy.dat">'
        '<code class="xref download docutils literal notranslate">'
        '<span class="pre">dummy.dat</span></code></a></p></li>' % digest
    ) in content
    assert (
        '<li><p><a class="reference download internal" download="" '
        'href="_downloads/%s/dummy.dat">'
        '<code class="xref download docutils literal notranslate">'
        '<span class="pre">another/dummy.dat</span></code></a></p></li>'
        % digest_another
    ) in content
    assert (
        '<li><p><code class="xref download docutils literal notranslate">'
        '<span class="pre">not_found.dat</span></code></p></li>'
    ) in content
    assert (
        '<li><p><a class="reference download external" download="" '
        'href="https://www.sphinx-doc.org/en/master/_static/sphinx-logo.svg">'
        '<code class="xref download docutils literal notranslate">'
        '<span class="pre">Sphinx</span> <span class="pre">logo</span>'
        '</code></a></p></li>'
    ) in content



"""Test the sphinx.quickstart module."""

from __future__ import annotations

import time
from io import StringIO
from typing import TYPE_CHECKING

import pytest

from sphinx._cli.util.colour import disable_colour, enable_colour
from sphinx.cmd import quickstart as qs
from sphinx.testing.util import SphinxTestApp

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path
    from typing import Any

warnfile = StringIO()


def setup_module():
    disable_colour()


def mock_input(
    answers: dict[str, str], needanswer: bool = False
) -> Callable[[str], str]:
    called = set()

    def input_(prompt: str) -> str:
        if prompt in called:
            raise AssertionError(
                'answer for %r missing and no default present' % prompt
            )
        called.add(prompt)
        for question, answer in answers.items():
            if prompt.startswith(qs.PROMPT_PREFIX + question):
                return answer
        if needanswer:
            raise AssertionError('answer for %r missing' % prompt)
        return ''

    return input_


real_input: Callable[[str], str] = input


def teardown_module():
    qs.term_input = real_input
    enable_colour()


def test_do_prompt() -> None:
    answers = {
        'Q2': 'v2',
        'Q3': 'v3',
        'Q4': 'yes',
        'Q5': 'no',
        'Q6': 'foo',
    }
    qs.term_input = mock_input(answers)  # type: ignore[assignment]

    assert qs.do_prompt('Q1', default='v1') == 'v1'
    assert qs.do_prompt('Q3', default='v3_default') == 'v3'
    assert qs.do_prompt('Q2') == 'v2'
    assert qs.do_prompt('Q4', validator=qs.boolean) is True
    assert qs.do_prompt('Q5', validator=qs.boolean) is False
    with pytest.raises(AssertionError):
        qs.do_prompt('Q6', validator=qs.boolean)


def test_do_prompt_inputstrip() -> None:
    answers = {
        'Q1': 'Y',
        'Q2': ' Yes ',
        'Q3': 'N',
        'Q4': 'N ',
    }
    qs.term_input = mock_input(answers)  # type: ignore[assignment]

    assert qs.do_prompt('Q1') == 'Y'
    assert qs.do_prompt('Q2') == 'Yes'
    assert qs.do_prompt('Q3') == 'N'
    assert qs.do_prompt('Q4') == 'N'


def test_do_prompt_with_nonascii() -> None:
    answers = {
        'Q1': '\u30c9\u30a4\u30c4',
    }
    qs.term_input = mock_input(answers)  # type: ignore[assignment]
    result = qs.do_prompt('Q1', default='\u65e5\u672c')
    assert result == '\u30c9\u30a4\u30c4'


def test_quickstart_defaults(tmp_path):
    answers = {
        'Root path': str(tmp_path),
        'Project name': 'Sphinx Test',
        'Author name': 'Georg Brandl',
        'Project version': '0.1',
    }
    qs.term_input = mock_input(answers)
    d: dict[str, Any] = {}
    qs.ask_user(d)
    qs.generate(d)

    conffile = tmp_path / 'conf.py'
    assert conffile.is_file()
    ns: dict[str, Any] = {}
    exec(conffile.read_text(encoding='utf8'), ns)  # NoQA: S102
    assert ns['extensions'] == []
    assert ns['templates_path'] == ['_templates']
    assert ns['project'] == 'Sphinx Test'
    assert ns['copyright'] == '%s, Georg Brandl' % time.strftime('%Y')
    assert ns['version'] == '0.1'
    assert ns['release'] == '0.1'
    assert ns['html_static_path'] == ['_static']

    assert (tmp_path / '_static').is_dir()
    assert (tmp_path / '_templates').is_dir()
    assert (tmp_path / 'index.rst').is_file()
    assert (tmp_path / 'Makefile').is_file()
    assert (tmp_path / 'make.bat').is_file()


def test_quickstart_all_answers(tmp_path):
    answers = {
        'Root path': str(tmp_path),
        'Separate source and build': 'y',
        'Name prefix for templates': '.',
        'Project name': 'STASI™',
        'Author name': "Wolfgang Schäuble & G'Beckstein",
        'Project version': '2.0',
        'Project release': '2.0.1',
        'Project language': 'de',
        'Source file suffix': '.txt',
        'Name of your master document': 'contents',
        'autodoc': 'y',
        'doctest': 'yes',
        'intersphinx': 'no',
        'todo': 'y',
        'coverage': 'no',
        'imgmath': 'N',
        'mathjax': 'no',
        'ifconfig': 'no',
        'viewcode': 'no',
        'githubpages': 'no',
        'Create Makefile': 'no',
        'Create Windows command file': 'no',
        'Do you want to use the epub builder': 'yes',
    }
    qs.term_input = mock_input(answers, needanswer=True)
    d: dict[str, Any] = {}
    qs.ask_user(d)
    qs.generate(d)

    conffile = tmp_path / 'source' / 'conf.py'
    assert conffile.is_file()
    ns: dict[str, Any] = {}
    exec(conffile.read_text(encoding='utf8'), ns)  # NoQA: S102
    assert ns['extensions'] == [
        'sphinx.ext.autodoc',
        'sphinx.ext.doctest',
        'sphinx.ext.todo',
    ]
    assert ns['templates_path'] == ['.templates']
    assert ns['source_suffix'] == '.txt'
    assert ns['root_doc'] == 'contents'
    assert ns['project'] == 'STASI™'
    assert ns['copyright'] == "%s, Wolfgang Schäuble & G'Beckstein" % time.strftime(
        '%Y'
    )
    assert ns['version'] == '2.0'
    assert ns['release'] == '2.0.1'
    assert ns['todo_include_todos'] is True
    assert ns['html_static_path'] == ['.static']

    assert (tmp_path / 'build').is_dir()
    assert (tmp_path / 'source' / '.static').is_dir()
    assert (tmp_path / 'source' / '.templates').is_dir()
    assert (tmp_path / 'source' / 'contents.txt').is_file()


def test_generated_files_eol(tmp_path):
    answers = {
        'Root path': str(tmp_path),
        'Project name': 'Sphinx Test',
        'Author name': 'Georg Brandl',
        'Project version': '0.1',
    }
    qs.term_input = mock_input(answers)
    d: dict[str, Any] = {}
    qs.ask_user(d)
    qs.generate(d)

    def assert_eol(filename: Path, eol: str) -> None:
        content = filename.read_bytes().decode()
        assert all(l[-len(eol) :] == eol for l in content.splitlines(keepends=True))

    assert_eol(tmp_path / 'make.bat', '\r\n')
    assert_eol(tmp_path / 'Makefile', '\n')


def test_quickstart_and_build(tmp_path):
    answers = {
        'Root path': str(tmp_path),
        'Project name': 'Fullwidth characters: \u30c9\u30a4\u30c4',
        'Author name': 'Georg Brandl',
        'Project version': '0.1',
    }
    qs.term_input = mock_input(answers)
    d: dict[str, Any] = {}
    qs.ask_user(d)
    qs.generate(d)

    app = SphinxTestApp('html', srcdir=tmp_path, warning=warnfile)
    app.build(force_all=True)
    app.cleanup()
    warnings = warnfile.getvalue()
    assert not warnings


def test_default_filename(tmp_path):
    answers = {
        'Root path': str(tmp_path),
        'Project name': '\u30c9\u30a4\u30c4',  # Fullwidth characters only
        'Author name': 'Georg Brandl',
        'Project version': '0.1',
    }
    qs.term_input = mock_input(answers)
    d: dict[str, Any] = {}
    qs.ask_user(d)
    qs.generate(d)

    conffile = tmp_path / 'conf.py'
    assert conffile.is_file()
    ns: dict[str, Any] = {}
    exec(conffile.read_text(encoding='utf8'), ns)  # NoQA: S102


def test_extensions(tmp_path):
    qs.main([
        '-q',
        '-p',
        'project_name',
        '-a',
        'author',
        '--extensions',
        'foo,bar,baz',
        str(tmp_path),
    ])

    conffile = tmp_path / 'conf.py'
    assert conffile.is_file()
    ns: dict[str, Any] = {}
    exec(conffile.read_text(encoding='utf8'), ns)  # NoQA: S102
    assert ns['extensions'] == ['foo', 'bar', 'baz']


def test_exits_when_existing_confpy(monkeypatch):
    # The code detects existing conf.py with path.is_file()
    # so we mock it as True with pytest's monkeypatch
    monkeypatch.setattr('os.path.isfile', lambda path: True)

    qs.term_input = mock_input({
        'Please enter a new root path (or just Enter to exit)': '',
    })
    d: dict[str, Any] = {}
    with pytest.raises(SystemExit):
        qs.ask_user(d)



from pathlib import Path

from sphinx.util.fileutil import copy_asset


def _copy_asset_overwrite_hook(app):
    css = app.outdir / '_static' / 'custom-styles.css'
    # html_static_path is copied by default
    css_content = css.read_text(encoding='utf-8')
    assert css_content == '/* html_static_path */\n', 'invalid default text'
    # warning generated by here
    copy_asset(
        Path(__file__).resolve().parent.joinpath('myext_static', 'custom-styles.css'),
        app.outdir / '_static',
    )
    # This demonstrates that no overwriting occurs
    css_content = css.read_text(encoding='utf-8')
    assert css_content == '/* html_static_path */\n', 'file overwritten!'
    return []


def setup(app):
    app.connect('html-collect-pages', _copy_asset_overwrite_hook)
    app.add_css_file('custom-styles.css')



"""Test sphinx.roles"""

from __future__ import annotations

from unittest.mock import Mock

import pytest
from docutils import nodes

from sphinx.roles import EmphasizedLiteral, _format_rfc_target
from sphinx.testing.util import assert_node


def test_samp() -> None:
    emph_literal_role = EmphasizedLiteral()

    # normal case
    text = 'print 1+{variable}'
    ret, msg = emph_literal_role('samp', text, text, 0, Mock())
    assert_node(
        ret[0],
        [
            nodes.literal,
            (
                'print 1+',
                [nodes.emphasis, 'variable'],
            ),
        ],
    )
    assert msg == []

    # two emphasis items
    text = 'print {1}+{variable}'
    ret, msg = emph_literal_role('samp', text, text, 0, Mock())
    assert_node(
        ret[0],
        [
            nodes.literal,
            (
                'print ',
                [nodes.emphasis, '1'],
                '+',
                [nodes.emphasis, 'variable'],
            ),
        ],
    )
    assert msg == []

    # empty curly brace
    text = 'print 1+{}'
    ret, msg = emph_literal_role('samp', text, text, 0, Mock())
    assert_node(ret[0], [nodes.literal, 'print 1+{}'])
    assert msg == []

    # half-opened variable
    text = 'print 1+{variable'
    ret, msg = emph_literal_role('samp', text, text, 0, Mock())
    assert_node(ret[0], [nodes.literal, 'print 1+{variable'])
    assert msg == []

    # nested
    text = 'print 1+{{variable}}'
    ret, msg = emph_literal_role('samp', text, text, 0, Mock())
    assert_node(
        ret[0],
        [
            nodes.literal,
            (
                'print 1+',
                [nodes.emphasis, '{variable'],
                '}',
            ),
        ],
    )
    assert msg == []

    # emphasized item only
    text = '{variable}'
    ret, msg = emph_literal_role('samp', text, text, 0, Mock())
    assert_node(ret[0], [nodes.literal, nodes.emphasis, 'variable'])
    assert msg == []

    # escaping
    text = r'print 1+\{variable}'
    ret, msg = emph_literal_role('samp', text, text, 0, Mock())
    assert_node(ret[0], [nodes.literal, 'print 1+{variable}'])
    assert msg == []

    # escaping (2)
    text = r'print 1+\{{variable}\}'
    ret, msg = emph_literal_role('samp', text, text, 0, Mock())
    assert_node(
        ret[0],
        [
            nodes.literal,
            (
                'print 1+{',
                [nodes.emphasis, 'variable'],
                '}',
            ),
        ],
    )
    assert msg == []

    # escape a backslash
    text = r'print 1+\\{variable}'
    ret, msg = emph_literal_role('samp', text, text, 0, Mock())
    assert_node(
        ret[0],
        [
            nodes.literal,
            (
                'print 1+\\',
                [nodes.emphasis, 'variable'],
            ),
        ],
    )
    assert msg == []


@pytest.mark.parametrize(
    ('target', 'expected_output'),
    [
        ('123', 'RFC 123'),
        ('123#', 'RFC 123#'),
        ('123#id1', 'RFC 123#id1'),
        ('123#section', 'RFC 123 Section'),
        ('123#section-1', 'RFC 123 Section 1'),
        ('123#section-2.5.3', 'RFC 123 Section 2.5.3'),
        ('123#page-13', 'RFC 123 Page 13'),
        ('123#appendix-B', 'RFC 123 Appendix B'),
        # Section names are also present in some RFC anchors. Until we come up with a reliable way
        # to reconstruct the section names from the corresponding anchors with the correct
        # capitalization it's probably better to leave them alone.
        ('9076#name-risks-in-the-dns-data', 'RFC 9076#name-risks-in-the-dns-data'),
    ],
)
def test_format_rfc_target(target: str, expected_output: str) -> None:
    assert _format_rfc_target(target) == expected_output



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='trim_doctest_flags')
def test_trim_doctest_flags_html(app: SphinxTestApp) -> None:
    app.build()

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert 'FOO' not in result
    assert 'BAR' in result
    assert 'BAZ' not in result
    assert 'QUX' not in result
    assert 'QUUX' not in result
    assert 'CORGE' not in result
    assert 'GRAULT' in result
    assert '<span class="n">now</span><span class="p">()</span>   \n' not in result
    assert '<span class="n">now</span><span class="p">()</span>\n' in result


@pytest.mark.sphinx(
    'html',
    testroot='trim_doctest_flags',
    confoverrides={'trim_doctest_flags': False},
)
def test_trim_doctest_flags_disabled(app: SphinxTestApp) -> None:
    app.build()

    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert 'FOO' in result
    assert 'BAR' in result
    assert 'BAZ' in result
    assert 'QUX' in result
    assert 'QUUX' not in result
    assert 'CORGE' not in result
    assert 'GRAULT' in result


@pytest.mark.sphinx('latex', testroot='trim_doctest_flags')
def test_trim_doctest_flags_latex(app: SphinxTestApp) -> None:
    app.build()

    result = (app.outdir / 'projectnamenotset.tex').read_text(encoding='utf8')
    assert 'FOO' not in result
    assert 'BAR' in result
    assert 'BAZ' not in result
    assert 'QUX' not in result
    assert 'QUUX' not in result
    assert 'CORGE' not in result
    assert 'GRAULT' in result



"""Test sphinx.ext.todo extension."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx(
    'html',
    testroot='ext-todo',
    freshenv=True,
    confoverrides={'todo_include_todos': True, 'todo_emit_warnings': True},
)
def test_todo(app: SphinxTestApp) -> None:
    todos = []

    def on_todo_defined(app, node):
        todos.append(node)

    app.connect('todo-defined', on_todo_defined)
    app.build(force_all=True)

    # check todolist
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<p class="admonition-title">Todo</p>\n<p>todo in foo</p>' in content

    assert '<p class="admonition-title">Todo</p>\n<p>todo in bar</p>' in content

    # check todo
    content = (app.outdir / 'foo.html').read_text(encoding='utf8')
    assert '<p class="admonition-title">Todo</p>\n<p>todo in foo</p>' in content

    assert (
        '<p class="admonition-title">Todo</p>\n<p>todo in param field</p>'
    ) in content

    # check emitted warnings
    assert 'WARNING: TODO entry found: todo in foo' in app.warning.getvalue()
    assert 'WARNING: TODO entry found: todo in bar' in app.warning.getvalue()

    # check handled event
    assert len(todos) == 3
    assert {todo[1].astext() for todo in todos} == {
        'todo in foo',
        'todo in bar',
        'todo in param field',
    }


@pytest.mark.sphinx(
    'html',
    testroot='ext-todo',
    freshenv=True,
    confoverrides={'todo_include_todos': False, 'todo_emit_warnings': True},
)
def test_todo_not_included(app: SphinxTestApp) -> None:
    todos = []

    def on_todo_defined(app, node):
        todos.append(node)

    app.connect('todo-defined', on_todo_defined)
    app.build(force_all=True)

    # check todolist
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert '<p class="admonition-title">Todo</p>\n<p>todo in foo</p>' not in content

    assert '<p class="admonition-title">Todo</p>\n<p>todo in bar</p>' not in content

    # check todo
    content = (app.outdir / 'foo.html').read_text(encoding='utf8')
    assert '<p class="admonition-title">Todo</p>\n<p>todo in foo</p>' not in content

    # check emitted warnings
    assert 'WARNING: TODO entry found: todo in foo' in app.warning.getvalue()
    assert 'WARNING: TODO entry found: todo in bar' in app.warning.getvalue()

    # check handled event
    assert len(todos) == 3
    assert {todo[1].astext() for todo in todos} == {
        'todo in foo',
        'todo in bar',
        'todo in param field',
    }


@pytest.mark.sphinx(
    'latex',
    testroot='ext-todo',
    freshenv=True,
    confoverrides={'todo_include_todos': True},
)
def test_todo_valid_link(app: SphinxTestApp) -> None:
    """Test that the inserted "original entry" links for todo items have a target
    that exists in the LaTeX output. The target was previously incorrectly omitted.
    https://github.com/sphinx-doc/sphinx/issues/1020
    """
    # Ensure the LaTeX output is built.
    app.build(force_all=True)

    content = (app.outdir / 'projectnamenotset.tex').read_text(encoding='utf8')

    # Look for the link to foo. Note that there are two of them because the
    # source document uses todolist twice. We could equally well look for links
    # to bar.
    link = (
        r'{\\hyperref\[\\detokenize{(.*?foo.*?)}]{\\sphinxcrossref{'
        r'\\sphinxstyleemphasis{original entry}}}}'
    )
    m = re.findall(link, content)
    assert len(m) == 4
    target = m[0]

    # Look for the targets of this link.
    labels = re.findall(r'\\label{\\detokenize{([^}]*)}}', content)
    matched = [l for l in labels if l == target]

    # If everything is correct we should have exactly one target.
    assert len(matched) == 1



# from __future__ import annotations

import dataclasses
import types
from typing import Annotated


@dataclasses.dataclass(frozen=True)
class FuncValidator:
    func: types.FunctionType


@dataclasses.dataclass(frozen=True)
class MaxLen:
    max_length: int
    whitelisted_words: list[str]


def validate(value: str) -> str:
    return value


#: Type alias for a validated string.
ValidatedString = Annotated[str, FuncValidator(validate)]


def hello(name: Annotated[str, 'attribute']) -> None:
    """docstring"""
    pass


class AnnotatedAttributes:
    """docstring"""

    #: Docstring about the ``name`` attribute.
    name: Annotated[str, 'attribute']

    #: Docstring about the ``max_len`` attribute.
    max_len: list[Annotated[str, MaxLen(10, ['word_one', 'word_two'])]]

    #: Docstring about the ``validated`` attribute.
    validated: ValidatedString



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx(
    'html',
    testroot='ext-extlinks-hardcoded-urls',
    confoverrides={'extlinks_detect_hardcoded_links': False},
)
def test_extlinks_detect_candidates(app: SphinxTestApp) -> None:
    app.build()
    assert app.warning.getvalue() == ''


@pytest.mark.sphinx('html', testroot='ext-extlinks-hardcoded-urls')
def test_replaceable_uris_emit_extlinks_warnings(app: SphinxTestApp) -> None:
    app.build()
    warning_output = app.warning.getvalue()

    # there should be exactly three warnings for replaceable URLs
    message = (
        "index.rst:%d: WARNING: hardcoded link 'https://github.com/sphinx-doc/sphinx/issues/1' "
        "could be replaced by an extlink (try using '%s' instead)"
    )
    assert message % (11, ':issue:`1`') in warning_output
    assert message % (13, ':issue:`inline replaceable link <1>`') in warning_output
    assert message % (15, ':issue:`replaceable link <1>`') in warning_output


@pytest.mark.sphinx(
    'html',
    testroot='ext-extlinks-hardcoded-urls-multiple-replacements',
)
def test_all_replacements_suggested_if_multiple_replacements_possible(
    app: SphinxTestApp,
) -> None:
    app.build()
    warning_output = app.warning.getvalue()
    # there should be six warnings for replaceable URLs, three pairs per link
    assert warning_output.count('WARNING: hardcoded link') == 6
    message = (
        "index.rst:%d: WARNING: hardcoded link 'https://github.com/octocat' "
        "could be replaced by an extlink (try using '%s' instead)"
    )
    assert message % (14, ':user:`octocat`') in warning_output
    assert message % (16, ':user:`inline replaceable link <octocat>`') in warning_output
    assert message % (18, ':user:`replaceable link <octocat>`') in warning_output
    message = (
        "index.rst:%d: WARNING: hardcoded link 'https://github.com/octocat' "
        "could be replaced by an extlink (try using '%s' instead)"
    )
    assert message % (14, ':repo:`octocat`') in warning_output
    assert message % (16, ':repo:`inline replaceable link <octocat>`') in warning_output
    assert message % (18, ':repo:`replaceable link <octocat>`') in warning_output



"""Tests the Python Domain"""

from __future__ import annotations

import re
from unittest.mock import Mock

import docutils.utils
import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.addnodes import (
    desc,
    desc_annotation,
    desc_content,
    desc_name,
    desc_parameter,
    desc_parameterlist,
    desc_returns,
    desc_sig_keyword,
    desc_sig_literal_number,
    desc_sig_literal_string,
    desc_sig_name,
    desc_sig_operator,
    desc_sig_punctuation,
    desc_sig_space,
    desc_signature,
    desc_type_parameter,
    desc_type_parameter_list,
    pending_xref,
)
from sphinx.domains import IndexEntry
from sphinx.domains.python import PythonDomain, PythonModuleIndex
from sphinx.domains.python._annotations import _parse_annotation, _pseudo_parse_arglist
from sphinx.domains.python._object import py_sig_re
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node
from sphinx.writers.text import STDINDENT


def parse(sig):
    m = py_sig_re.match(sig)
    if m is None:
        raise ValueError
    name_prefix, tp_list, name, arglist, retann = m.groups()
    signode = addnodes.desc_signature(sig, '')
    _pseudo_parse_arglist(signode, arglist)
    return signode.astext()


def test_function_signatures() -> None:
    rv = parse("compile(source : string, filename, symbol='file')")
    assert rv == "(source : string, filename, symbol='file')"

    for params, expect in [
        ('(a=1)', '(a=1)'),
        ('(a: int = 1)', '(a: int = 1)'),
        ('(a=1, [b=None])', '(a=1, [b=None])'),
        ('(a=1[, b=None])', '(a=1, [b=None])'),
        ('(a=[], [b=None])', '(a=[], [b=None])'),
        ('(a=[][, b=None])', '(a=[], [b=None])'),
        ('(a: Foo[Bar]=[][, b=None])', '(a: Foo[Bar]=[], [b=None])'),
    ]:
        rv = parse(f'func{params}')
        assert rv == expect

        # Note: 'def f[Foo[Bar]]()' is not valid Python but people might write
        # it in a reST document to convene the intent of a higher-kinded type
        # variable.
        for tparams in ['', '[Foo]', '[Foo[Bar]]']:
            for retann in ['', '-> Foo', '-> Foo[Bar]', '-> anything else']:
                rv = parse(f'func{tparams}{params} {retann}'.rstrip())
                assert rv == expect


@pytest.mark.sphinx('dummy', testroot='domain-py')
def test_domain_py_xrefs(app):
    """Domain objects have correct prefixes when looking up xrefs"""
    app.build(force_all=True)

    def assert_refnode(
        node, module_name, class_name, target, reftype=None, domain='py'
    ):
        attributes = {
            'refdomain': domain,
            'reftarget': target,
        }
        if reftype is not None:
            attributes['reftype'] = reftype
        if module_name is not False:
            attributes['py:module'] = module_name
        if class_name is not False:
            attributes['py:class'] = class_name
        assert_node(node, **attributes)

    doctree = app.env.get_doctree('roles')
    refnodes = list(doctree.findall(pending_xref))
    assert_refnode(refnodes[0], None, None, 'TopLevel', 'class')
    assert_refnode(refnodes[1], None, None, 'top_level', 'meth')
    assert_refnode(refnodes[2], None, None, 'TopLevelType', 'type')
    assert_refnode(refnodes[3], None, 'NestedParentA', 'child_1', 'meth')
    assert_refnode(
        refnodes[4], None, 'NestedParentA', 'NestedChildA.subchild_2', 'meth'
    )
    assert_refnode(refnodes[5], None, 'NestedParentA', 'child_2', 'meth')
    assert_refnode(refnodes[6], False, 'NestedParentA', 'any_child', domain='')
    assert_refnode(refnodes[7], None, 'NestedParentA', 'NestedChildA', 'class')
    assert_refnode(
        refnodes[8], None, 'NestedParentA.NestedChildA', 'subchild_2', 'meth'
    )
    assert_refnode(
        refnodes[9], None, 'NestedParentA.NestedChildA', 'NestedParentA.child_1', 'meth'
    )
    assert_refnode(
        refnodes[10], None, 'NestedParentA', 'NestedChildA.subchild_1', 'meth'
    )
    assert_refnode(refnodes[11], None, 'NestedParentB', 'child_1', 'meth')
    assert_refnode(refnodes[12], None, 'NestedParentB', 'NestedParentB', 'class')
    assert_refnode(refnodes[13], None, None, 'NestedParentA.NestedChildA', 'class')
    assert_refnode(refnodes[14], None, None, 'NestedParentA.NestedTypeA', 'type')
    assert len(refnodes) == 15

    doctree = app.env.get_doctree('module')
    refnodes = list(doctree.findall(pending_xref))
    assert_refnode(refnodes[0], 'module_a.submodule', None, 'ModTopLevel', 'class')
    assert_refnode(
        refnodes[1], 'module_a.submodule', 'ModTopLevel', 'mod_child_1', 'meth'
    )
    assert_refnode(
        refnodes[2],
        'module_a.submodule',
        'ModTopLevel',
        'ModTopLevel.mod_child_1',
        'meth',
    )
    assert_refnode(
        refnodes[3], 'module_a.submodule', 'ModTopLevel', 'mod_child_2', 'meth'
    )
    assert_refnode(
        refnodes[4],
        'module_a.submodule',
        'ModTopLevel',
        'module_a.submodule.ModTopLevel.mod_child_1',
        'meth',
    )
    assert_refnode(refnodes[5], 'module_a.submodule', 'ModTopLevel', 'prop', 'attr')
    assert_refnode(refnodes[6], 'module_a.submodule', 'ModTopLevel', 'prop', 'meth')
    assert_refnode(refnodes[7], 'module_b.submodule', None, 'ModTopLevel', 'class')
    assert_refnode(
        refnodes[8], 'module_b.submodule', 'ModTopLevel', 'ModNoModule', 'class'
    )
    assert_refnode(refnodes[9], False, False, 'int', 'class')
    assert_refnode(refnodes[10], False, False, 'tuple', 'class')
    assert_refnode(refnodes[11], False, False, 'str', 'class')
    assert_refnode(refnodes[12], False, False, 'float', 'class')
    assert_refnode(refnodes[13], False, False, 'list', 'class')
    assert_refnode(refnodes[14], False, False, 'ModTopLevel', 'class')
    assert_refnode(refnodes[15], False, False, 'index', 'doc', domain='std')
    assert_refnode(refnodes[16], False, False, 'typing.Literal', 'obj', domain='py')
    assert_refnode(refnodes[17], False, False, 'typing.Literal', 'obj', domain='py')
    assert_refnode(refnodes[18], False, False, 'list', 'class', domain='py')
    assert_refnode(refnodes[19], False, False, 'int', 'class', domain='py')
    assert_refnode(refnodes[20], False, False, 'str', 'class', domain='py')
    assert len(refnodes) == 21

    doctree = app.env.get_doctree('module_option')
    refnodes = list(doctree.findall(pending_xref))
    print(refnodes)
    print(refnodes[0])
    print(refnodes[1])
    assert_refnode(refnodes[0], 'test.extra', 'B', 'foo', 'meth')
    assert_refnode(refnodes[1], 'test.extra', 'B', 'foo', 'meth')
    assert len(refnodes) == 2


@pytest.mark.sphinx('html', testroot='domain-py')
def test_domain_py_xrefs_abbreviations(app):
    app.build(force_all=True)

    content = (app.outdir / 'abbr.html').read_text(encoding='utf8')
    assert re.search(
        r'normal: <a .* href="module.html#module_a.submodule.ModTopLevel.'
        r'mod_child_1" .*><.*>module_a.submodule.ModTopLevel.mod_child_1\(\)'
        r'<.*></a>',
        content,
    )
    assert re.search(
        r'relative: <a .* href="module.html#module_a.submodule.ModTopLevel.'
        r'mod_child_1" .*><.*>ModTopLevel.mod_child_1\(\)<.*></a>',
        content,
    )
    assert re.search(
        r'short name: <a .* href="module.html#module_a.submodule.ModTopLevel.'
        r'mod_child_1" .*><.*>mod_child_1\(\)<.*></a>',
        content,
    )
    assert re.search(
        r'relative \+ short name: <a .* href="module.html#module_a.submodule.'
        r'ModTopLevel.mod_child_1" .*><.*>mod_child_1\(\)<.*></a>',
        content,
    )
    assert re.search(
        r'short name \+ relative: <a .* href="module.html#module_a.submodule.'
        r'ModTopLevel.mod_child_1" .*><.*>mod_child_1\(\)<.*></a>',
        content,
    )


@pytest.mark.sphinx('dummy', testroot='domain-py')
def test_domain_py_objects(app):
    app.build(force_all=True)

    modules = app.env.domains.python_domain.data['modules']
    objects = app.env.domains.python_domain.data['objects']

    assert 'module_a.submodule' in modules
    assert 'module_a.submodule' in objects
    assert 'module_b.submodule' in modules
    assert 'module_b.submodule' in objects

    assert objects['module_a.submodule.ModTopLevel'][2] == 'class'
    assert objects['module_a.submodule.ModTopLevel.mod_child_1'][2] == 'method'
    assert objects['module_a.submodule.ModTopLevel.mod_child_2'][2] == 'method'
    assert 'ModTopLevel.ModNoModule' not in objects
    assert objects['ModNoModule'][2] == 'class'
    assert objects['module_b.submodule.ModTopLevel'][2] == 'class'

    assert objects['TopLevel'][2] == 'class'
    assert objects['top_level'][2] == 'method'
    assert objects['TopLevelType'][2] == 'type'
    assert objects['NestedParentA'][2] == 'class'
    assert objects['NestedParentA.NestedTypeA'][2] == 'type'
    assert objects['NestedParentA.child_1'][2] == 'method'
    assert objects['NestedParentA.any_child'][2] == 'method'
    assert objects['NestedParentA.NestedChildA'][2] == 'class'
    assert objects['NestedParentA.NestedChildA.subchild_1'][2] == 'method'
    assert objects['NestedParentA.NestedChildA.subchild_2'][2] == 'method'
    assert objects['NestedParentA.child_2'][2] == 'method'
    assert objects['NestedParentB'][2] == 'class'
    assert objects['NestedParentB.child_1'][2] == 'method'


@pytest.mark.sphinx('html', testroot='domain-py')
def test_resolve_xref_for_properties(app):
    app.build(force_all=True)

    content = (app.outdir / 'module.html').read_text(encoding='utf8')
    assert (
        'Link to <a class="reference internal" href="#module_a.submodule.ModTopLevel.prop"'
        ' title="module_a.submodule.ModTopLevel.prop">'
        '<code class="xref py py-attr docutils literal notranslate"><span class="pre">'
        'prop</span> <span class="pre">attribute</span></code></a>'
    ) in content
    assert (
        'Link to <a class="reference internal" href="#module_a.submodule.ModTopLevel.prop"'
        ' title="module_a.submodule.ModTopLevel.prop">'
        '<code class="xref py py-meth docutils literal notranslate"><span class="pre">'
        'prop</span> <span class="pre">method</span></code></a>'
    ) in content
    assert (
        'Link to <a class="reference internal" href="#module_a.submodule.ModTopLevel.prop"'
        ' title="module_a.submodule.ModTopLevel.prop">'
        '<code class="xref py py-attr docutils literal notranslate"><span class="pre">'
        'prop</span> <span class="pre">attribute</span></code></a>'
    ) in content


@pytest.mark.sphinx('dummy', testroot='domain-py')
def test_domain_py_find_obj(app):
    def find_obj(modname, prefix, obj_name, obj_type, searchmode=0):
        return app.env.domains.python_domain.find_obj(
            app.env, modname, prefix, obj_name, obj_type, searchmode
        )

    app.build(force_all=True)

    assert find_obj(None, None, 'NONEXISTANT', 'class') == []
    assert find_obj(None, None, 'NestedParentA', 'class') == [
        (
            'NestedParentA',
            ('roles', 'NestedParentA', 'class', False),
        )
    ]
    assert find_obj(None, None, 'NestedParentA.NestedTypeA', 'type') == [
        (
            'NestedParentA.NestedTypeA',
            ('roles', 'NestedParentA.NestedTypeA', 'type', False),
        )
    ]
    assert find_obj(None, None, 'NestedParentA.NestedChildA', 'class') == [
        (
            'NestedParentA.NestedChildA',
            ('roles', 'NestedParentA.NestedChildA', 'class', False),
        )
    ]
    assert find_obj(None, 'NestedParentA', 'NestedChildA', 'class') == [
        (
            'NestedParentA.NestedChildA',
            ('roles', 'NestedParentA.NestedChildA', 'class', False),
        )
    ]
    assert find_obj(None, None, 'NestedParentA.NestedChildA.subchild_1', 'meth') == [
        (
            'NestedParentA.NestedChildA.subchild_1',
            ('roles', 'NestedParentA.NestedChildA.subchild_1', 'method', False),
        )
    ]
    assert find_obj(None, 'NestedParentA', 'NestedChildA.subchild_1', 'meth') == [
        (
            'NestedParentA.NestedChildA.subchild_1',
            ('roles', 'NestedParentA.NestedChildA.subchild_1', 'method', False),
        )
    ]
    assert find_obj(None, 'NestedParentA.NestedChildA', 'subchild_1', 'meth') == [
        (
            'NestedParentA.NestedChildA.subchild_1',
            ('roles', 'NestedParentA.NestedChildA.subchild_1', 'method', False),
        )
    ]


@pytest.mark.sphinx('html', testroot='root')
def test_get_full_qualified_name() -> None:
    env = Mock(domaindata={})
    domain = PythonDomain(env)

    # non-python references
    node = nodes.reference()
    assert domain.get_full_qualified_name(node) is None

    # simple reference
    node = nodes.reference(reftarget='func')
    assert domain.get_full_qualified_name(node) == 'func'

    # with py:module context
    kwargs = {'py:module': 'module1'}
    node = nodes.reference(reftarget='func', **kwargs)
    assert domain.get_full_qualified_name(node) == 'module1.func'

    # with py:class context
    kwargs = {'py:class': 'Class'}
    node = nodes.reference(reftarget='func', **kwargs)
    assert domain.get_full_qualified_name(node) == 'Class.func'

    # with both py:module and py:class context
    kwargs = {'py:module': 'module1', 'py:class': 'Class'}
    node = nodes.reference(reftarget='func', **kwargs)
    assert domain.get_full_qualified_name(node) == 'module1.Class.func'


@pytest.mark.sphinx('html', testroot='_blank')
def test_parse_annotation(app):
    doctree = _parse_annotation('int', app.env)
    assert_node(doctree, ([pending_xref, 'int'],))
    assert_node(
        doctree[0], pending_xref, refdomain='py', reftype='class', reftarget='int'
    )

    doctree = _parse_annotation('List[int]', app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'List'],
            [desc_sig_punctuation, '['],
            [pending_xref, 'int'],
            [desc_sig_punctuation, ']'],
        ),
    )

    doctree = _parse_annotation('Tuple[int, int]', app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'Tuple'],
            [desc_sig_punctuation, '['],
            [pending_xref, 'int'],
            [desc_sig_punctuation, ','],
            desc_sig_space,
            [pending_xref, 'int'],
            [desc_sig_punctuation, ']'],
        ),
    )

    doctree = _parse_annotation('Tuple[()]', app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'Tuple'],
            [desc_sig_punctuation, '['],
            [desc_sig_punctuation, '('],
            [desc_sig_punctuation, ')'],
            [desc_sig_punctuation, ']'],
        ),
    )

    doctree = _parse_annotation('Tuple[int, ...]', app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'Tuple'],
            [desc_sig_punctuation, '['],
            [pending_xref, 'int'],
            [desc_sig_punctuation, ','],
            desc_sig_space,
            [desc_sig_punctuation, '...'],
            [desc_sig_punctuation, ']'],
        ),
    )

    doctree = _parse_annotation('Callable[[int, int], int]', app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'Callable'],
            [desc_sig_punctuation, '['],
            [desc_sig_punctuation, '['],
            [pending_xref, 'int'],
            [desc_sig_punctuation, ','],
            desc_sig_space,
            [pending_xref, 'int'],
            [desc_sig_punctuation, ']'],
            [desc_sig_punctuation, ','],
            desc_sig_space,
            [pending_xref, 'int'],
            [desc_sig_punctuation, ']'],
        ),
    )

    doctree = _parse_annotation('Callable[[], int]', app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'Callable'],
            [desc_sig_punctuation, '['],
            [desc_sig_punctuation, '['],
            [desc_sig_punctuation, ']'],
            [desc_sig_punctuation, ','],
            desc_sig_space,
            [pending_xref, 'int'],
            [desc_sig_punctuation, ']'],
        ),
    )

    doctree = _parse_annotation('List[None]', app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'List'],
            [desc_sig_punctuation, '['],
            [pending_xref, 'None'],
            [desc_sig_punctuation, ']'],
        ),
    )

    # None type makes an object-reference (not a class reference)
    doctree = _parse_annotation('None', app.env)
    assert_node(doctree, ([pending_xref, 'None'],))
    assert_node(
        doctree[0], pending_xref, refdomain='py', reftype='obj', reftarget='None'
    )

    # Literal type makes an object-reference (not a class reference)
    doctree = _parse_annotation("typing.Literal['a', 'b']", app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'Literal'],
            [desc_sig_punctuation, '['],
            [desc_sig_literal_string, "'a'"],
            [desc_sig_punctuation, ','],
            desc_sig_space,
            [desc_sig_literal_string, "'b'"],
            [desc_sig_punctuation, ']'],
        ),
    )
    assert_node(
        doctree[0],
        pending_xref,
        refdomain='py',
        reftype='obj',
        reftarget='typing.Literal',
    )

    # Annotated type with callable gets parsed
    doctree = _parse_annotation(
        'Annotated[Optional[str], annotated_types.MaxLen(max_length=10)]', app.env
    )
    assert_node(
        doctree,
        (
            [pending_xref, 'Annotated'],
            [desc_sig_punctuation, '['],
            [pending_xref, 'str'],
            [desc_sig_space, ' '],
            [desc_sig_punctuation, '|'],
            [desc_sig_space, ' '],
            [pending_xref, 'None'],
            [desc_sig_punctuation, ','],
            [desc_sig_space, ' '],
            [pending_xref, 'annotated_types.MaxLen'],
            [desc_sig_punctuation, '('],
            [desc_sig_name, 'max_length'],
            [desc_sig_operator, '='],
            [desc_sig_literal_number, '10'],
            [desc_sig_punctuation, ')'],
            [desc_sig_punctuation, ']'],
        ),
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_parse_annotation_suppress(app):
    doctree = _parse_annotation('~typing.Dict[str, str]', app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'Dict'],
            [desc_sig_punctuation, '['],
            [pending_xref, 'str'],
            [desc_sig_punctuation, ','],
            desc_sig_space,
            [pending_xref, 'str'],
            [desc_sig_punctuation, ']'],
        ),
    )
    assert_node(
        doctree[0], pending_xref, refdomain='py', reftype='obj', reftarget='typing.Dict'
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_parse_annotation_Literal(app):
    doctree = _parse_annotation('Literal[True, False]', app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'Literal'],
            [desc_sig_punctuation, '['],
            [desc_sig_keyword, 'True'],
            [desc_sig_punctuation, ','],
            desc_sig_space,
            [desc_sig_keyword, 'False'],
            [desc_sig_punctuation, ']'],
        ),
    )

    doctree = _parse_annotation("typing.Literal[0, 1, 'abc']", app.env)
    assert_node(
        doctree,
        (
            [pending_xref, 'Literal'],
            [desc_sig_punctuation, '['],
            [desc_sig_literal_number, '0'],
            [desc_sig_punctuation, ','],
            desc_sig_space,
            [desc_sig_literal_number, '1'],
            [desc_sig_punctuation, ','],
            desc_sig_space,
            [desc_sig_literal_string, "'abc'"],
            [desc_sig_punctuation, ']'],
        ),
    )


@pytest.mark.sphinx('html', testroot='root', freshenv=True)
def test_module_index(app):
    text = (
        '.. py:module:: docutils\n'
        '.. py:module:: sphinx\n'
        '.. py:module:: sphinx.config\n'
        '.. py:module:: sphinx.builders\n'
        '.. py:module:: sphinx.builders.html\n'
        '.. py:module:: sphinx_intl\n'
    )
    restructuredtext.parse(app, text)
    index = PythonModuleIndex(app.env.domains.python_domain)
    assert index.generate() == (
        [
            (
                'd',
                [
                    IndexEntry(
                        name='docutils',
                        subtype=0,
                        docname='index',
                        anchor='module-docutils',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                ],
            ),
            (
                's',
                [
                    IndexEntry(
                        name='sphinx',
                        subtype=1,
                        docname='index',
                        anchor='module-sphinx',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                    IndexEntry(
                        name='sphinx.builders',
                        subtype=2,
                        docname='index',
                        anchor='module-sphinx.builders',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                    IndexEntry(
                        name='sphinx.builders.html',
                        subtype=2,
                        docname='index',
                        anchor='module-sphinx.builders.html',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                    IndexEntry(
                        name='sphinx.config',
                        subtype=2,
                        docname='index',
                        anchor='module-sphinx.config',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                    IndexEntry(
                        name='sphinx_intl',
                        subtype=0,
                        docname='index',
                        anchor='module-sphinx_intl',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                ],
            ),
        ],
        False,
    )


@pytest.mark.sphinx('html', testroot='root', freshenv=True)
def test_module_index_submodule(app):
    text = '.. py:module:: sphinx.config\n'
    restructuredtext.parse(app, text)
    index = PythonModuleIndex(app.env.domains.python_domain)
    assert index.generate() == (
        [
            (
                's',
                [
                    IndexEntry(
                        name='sphinx',
                        subtype=1,
                        docname='',
                        anchor='',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                    IndexEntry(
                        name='sphinx.config',
                        subtype=2,
                        docname='index',
                        anchor='module-sphinx.config',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                ],
            )
        ],
        False,
    )


@pytest.mark.sphinx('html', testroot='root', freshenv=True)
def test_module_index_not_collapsed(app):
    text = '.. py:module:: docutils\n.. py:module:: sphinx\n'
    restructuredtext.parse(app, text)
    index = PythonModuleIndex(app.env.domains.python_domain)
    assert index.generate() == (
        [
            (
                'd',
                [
                    IndexEntry(
                        name='docutils',
                        subtype=0,
                        docname='index',
                        anchor='module-docutils',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                ],
            ),
            (
                's',
                [
                    IndexEntry(
                        name='sphinx',
                        subtype=0,
                        docname='index',
                        anchor='module-sphinx',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                ],
            ),
        ],
        True,
    )


@pytest.mark.sphinx(
    'html',
    testroot='root',
    freshenv=True,
    confoverrides={'modindex_common_prefix': ['sphinx.']},
)
def test_modindex_common_prefix(app):
    text = (
        '.. py:module:: docutils\n'
        '.. py:module:: sphinx\n'
        '.. py:module:: sphinx.config\n'
        '.. py:module:: sphinx.builders\n'
        '.. py:module:: sphinx.builders.html\n'
        '.. py:module:: sphinx_intl\n'
    )
    restructuredtext.parse(app, text)
    index = PythonModuleIndex(app.env.domains.python_domain)
    assert index.generate() == (
        [
            (
                'b',
                [
                    IndexEntry(
                        name='sphinx.builders',
                        subtype=1,
                        docname='index',
                        anchor='module-sphinx.builders',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                    IndexEntry(
                        name='sphinx.builders.html',
                        subtype=2,
                        docname='index',
                        anchor='module-sphinx.builders.html',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                ],
            ),
            (
                'c',
                [
                    IndexEntry(
                        name='sphinx.config',
                        subtype=0,
                        docname='index',
                        anchor='module-sphinx.config',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                ],
            ),
            (
                'd',
                [
                    IndexEntry(
                        name='docutils',
                        subtype=0,
                        docname='index',
                        anchor='module-docutils',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                ],
            ),
            (
                's',
                [
                    IndexEntry(
                        name='sphinx',
                        subtype=0,
                        docname='index',
                        anchor='module-sphinx',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                    IndexEntry(
                        name='sphinx_intl',
                        subtype=0,
                        docname='index',
                        anchor='module-sphinx_intl',
                        extra='',
                        qualifier='',
                        descr='',
                    ),
                ],
            ),
        ],
        True,
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_no_index_entry(app):
    text = '.. py:function:: f()\n.. py:function:: g()\n   :no-index-entry:\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, desc, addnodes.index, desc))
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('pair', 'built-in function; f()', 'f', '', None)],
    )
    assert_node(doctree[2], addnodes.index, entries=[])

    text = '.. py:class:: f\n.. py:class:: g\n   :no-index-entry:\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, desc, addnodes.index, desc))
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('single', 'f (built-in class)', 'f', '', None)],
    )
    assert_node(doctree[2], addnodes.index, entries=[])

    text = '.. py:module:: f\n.. py:module:: g\n   :no-index-entry:\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, nodes.target, nodes.target))
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('pair', 'module; f', 'module-f', '', None)],
    )


@pytest.mark.sphinx('html', testroot='domain-py-python_use_unqualified_type_names')
def test_python_python_use_unqualified_type_names(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<span class="n"><a class="reference internal" href="#foo.Name" title="foo.Name">'
        '<span class="pre">Name</span></a></span>'
    ) in content
    assert '<span class="n"><span class="pre">foo.Age</span></span>' in content
    assert (
        '<p><strong>name</strong> (<a class="reference internal" href="#foo.Name" '
        'title="foo.Name"><em>Name</em></a>) – blah blah</p>'
    ) in content
    assert '<p><strong>age</strong> (<em>foo.Age</em>) – blah blah</p>' in content


@pytest.mark.sphinx(
    'html',
    testroot='domain-py-python_use_unqualified_type_names',
    confoverrides={'python_use_unqualified_type_names': False},
)
def test_python_python_use_unqualified_type_names_disabled(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert (
        '<span class="n"><a class="reference internal" href="#foo.Name" title="foo.Name">'
        '<span class="pre">foo.Name</span></a></span>'
    ) in content
    assert '<span class="n"><span class="pre">foo.Age</span></span>' in content
    assert (
        '<p><strong>name</strong> (<a class="reference internal" href="#foo.Name" '
        'title="foo.Name"><em>foo.Name</em></a>) – blah blah</p>'
    ) in content
    assert '<p><strong>age</strong> (<em>foo.Age</em>) – blah blah</p>' in content


@pytest.mark.sphinx('dummy', testroot='domain-py-xref-warning')
def test_warn_missing_reference(app):
    app.build()
    assert "index.rst:6: WARNING: undefined label: 'no-label'" in app.warning.getvalue()
    assert (
        'index.rst:6: WARNING: Failed to create a cross reference. '
        "A title or caption not found: 'existing-label'"
    ) in app.warning.getvalue()


@pytest.mark.parametrize('include_options', [True, False])
@pytest.mark.sphinx('html', testroot='root', confoverrides={'nitpicky': True})
def test_signature_line_number(app, include_options):
    text = '.. py:function:: foo(bar : string)\n' + (
        '   :no-index-entry:\n' if include_options else ''
    )
    doc = restructuredtext.parse(app, text)
    xrefs = list(doc.findall(condition=addnodes.pending_xref))
    assert len(xrefs) == 1
    source, line = docutils.utils.get_source_line(xrefs[0])
    assert 'index.rst' in source
    assert line == 1


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'python_maximum_signature_line_length': len('hello(name: str) -> str'),
        'maximum_signature_line_length': 1,
    },
)
def test_python_maximum_signature_line_length_overrides_global(app):
    text = '.. py:function:: hello(name: str) -> str'
    doctree = restructuredtext.parse(app, text)
    expected_doctree = (
        addnodes.index,
        [
            desc,
            (
                [
                    desc_signature,
                    (
                        [desc_name, 'hello'],
                        desc_parameterlist,
                        [desc_returns, pending_xref, 'str'],
                    ),
                ],
                desc_content,
            ),
        ],
    )
    assert_node(doctree, expected_doctree)
    assert_node(
        doctree[1],
        addnodes.desc,
        desctype='function',
        domain='py',
        objtype='function',
        no_index=False,
    )
    signame_node = [desc_sig_name, 'name']
    expected_sig = [
        desc_parameterlist,
        desc_parameter,
        (
            signame_node,
            [desc_sig_punctuation, ':'],
            desc_sig_space,
            [nodes.inline, pending_xref, 'str'],
        ),
    ]
    assert_node(doctree[1][0][1], expected_sig)
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=False)


@pytest.mark.sphinx(
    'html',
    testroot='domain-py-python_maximum_signature_line_length',
)
def test_domain_py_python_maximum_signature_line_length_in_html(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    expected_parameter_list_hello = """\

<dl>
<dd>\
<em class="sig-param">\
<span class="n"><span class="pre">name</span></span>\
<span class="p"><span class="pre">:</span></span>\
<span class="w"> </span>\
<span class="n"><span class="pre">str</span></span>\
</em>,\
</dd>
</dl>

<span class="sig-paren">)</span> \
<span class="sig-return">\
<span class="sig-return-icon">&#x2192;</span> \
<span class="sig-return-typehint"><span class="pre">str</span></span>\
</span>\
<a class="headerlink" href="#hello" title="Link to this definition">¶</a>\
</dt>\
"""
    assert expected_parameter_list_hello in content

    param_line_fmt = '<dd>{}</dd>\n'
    param_name_fmt = (
        '<em class="sig-param"><span class="n"><span class="pre">{}</span></span></em>'
    )
    optional_fmt = '<span class="optional">{}</span>'

    expected_a = param_line_fmt.format(
        optional_fmt.format('[')
        + param_name_fmt.format('a')
        + ','
        + optional_fmt.format('['),
    )
    assert expected_a in content

    expected_b = param_line_fmt.format(
        param_name_fmt.format('b')
        + ','
        + optional_fmt.format(']')
        + optional_fmt.format(']'),
    )
    assert expected_b in content

    expected_c = param_line_fmt.format(param_name_fmt.format('c') + ',')
    assert expected_c in content

    expected_d = param_line_fmt.format(
        param_name_fmt.format('d') + optional_fmt.format('[') + ','
    )
    assert expected_d in content

    expected_e = param_line_fmt.format(param_name_fmt.format('e') + ',')
    assert expected_e in content

    expected_f = param_line_fmt.format(
        param_name_fmt.format('f') + ',' + optional_fmt.format(']')
    )
    assert expected_f in content

    expected_parameter_list_foo = """\

<dl>
{}{}{}{}{}{}</dl>

<span class="sig-paren">)</span>\
<a class="headerlink" href="#foo" title="Link to this definition">¶</a>\
</dt>\
""".format(expected_a, expected_b, expected_c, expected_d, expected_e, expected_f)
    assert expected_parameter_list_foo in content


@pytest.mark.sphinx(
    'text',
    testroot='domain-py-python_maximum_signature_line_length',
)
def test_domain_py_python_maximum_signature_line_length_in_text(app):
    app.build()
    content = (app.outdir / 'index.txt').read_text(encoding='utf8')
    param_line_fmt = STDINDENT * ' ' + '{}\n'

    expected_parameter_list_hello = '(\n{}) -> str'.format(
        param_line_fmt.format('name: str,')
    )

    assert expected_parameter_list_hello in content

    expected_a = param_line_fmt.format('[a,[')
    assert expected_a in content

    expected_b = param_line_fmt.format('b,]]')
    assert expected_b in content

    expected_c = param_line_fmt.format('c,')
    assert expected_c in content

    expected_d = param_line_fmt.format('d[,')
    assert expected_d in content

    expected_e = param_line_fmt.format('e,')
    assert expected_e in content

    expected_f = param_line_fmt.format('f,]')
    assert expected_f in content

    expected_parameter_list_foo = '(\n{}{}{}{}{}{})'.format(
        expected_a,
        expected_b,
        expected_c,
        expected_d,
        expected_e,
        expected_f,
    )
    assert expected_parameter_list_foo in content


@pytest.mark.sphinx(
    'html',
    testroot='domain-py-python_maximum_signature_line_length',
    confoverrides={'python_trailing_comma_in_multi_line_signatures': False},
)
def test_domain_py_python_trailing_comma_in_multi_line_signatures_in_html(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    expected_parameter_list_hello = """\

<dl>
<dd>\
<em class="sig-param">\
<span class="n"><span class="pre">name</span></span>\
<span class="p"><span class="pre">:</span></span>\
<span class="w"> </span>\
<span class="n"><span class="pre">str</span></span>\
</em>\
</dd>
</dl>

<span class="sig-paren">)</span> \
<span class="sig-return">\
<span class="sig-return-icon">&#x2192;</span> \
<span class="sig-return-typehint"><span class="pre">str</span></span>\
</span>\
<a class="headerlink" href="#hello" title="Link to this definition">¶</a>\
</dt>\
"""
    assert expected_parameter_list_hello in content

    param_line_fmt = '<dd>{}</dd>\n'
    param_name_fmt = (
        '<em class="sig-param"><span class="n"><span class="pre">{}</span></span></em>'
    )
    optional_fmt = '<span class="optional">{}</span>'

    expected_a = param_line_fmt.format(
        optional_fmt.format('[')
        + param_name_fmt.format('a')
        + ','
        + optional_fmt.format('['),
    )
    assert expected_a in content

    expected_b = param_line_fmt.format(
        param_name_fmt.format('b')
        + ','
        + optional_fmt.format(']')
        + optional_fmt.format(']'),
    )
    assert expected_b in content

    expected_c = param_line_fmt.format(param_name_fmt.format('c') + ',')
    assert expected_c in content

    expected_d = param_line_fmt.format(
        param_name_fmt.format('d') + optional_fmt.format('[') + ','
    )
    assert expected_d in content

    expected_e = param_line_fmt.format(param_name_fmt.format('e') + ',')
    assert expected_e in content

    expected_f = param_line_fmt.format(
        param_name_fmt.format('f') + optional_fmt.format(']')
    )
    assert expected_f in content

    expected_parameter_list_foo = """\

<dl>
{}{}{}{}{}{}</dl>

<span class="sig-paren">)</span>\
<a class="headerlink" href="#foo" title="Link to this definition">¶</a>\
</dt>\
""".format(expected_a, expected_b, expected_c, expected_d, expected_e, expected_f)
    assert expected_parameter_list_foo in content


@pytest.mark.sphinx(
    'text',
    testroot='domain-py-python_maximum_signature_line_length',
    freshenv=True,
    confoverrides={'python_trailing_comma_in_multi_line_signatures': False},
)
def test_domain_py_python_trailing_comma_in_multi_line_signatures_in_text(app):
    app.build()
    content = (app.outdir / 'index.txt').read_text(encoding='utf8')
    param_line_fmt = STDINDENT * ' ' + '{}\n'

    expected_parameter_list_hello = '(\n{}) -> str'.format(
        param_line_fmt.format('name: str')
    )

    assert expected_parameter_list_hello in content

    expected_a = param_line_fmt.format('[a,[')
    assert expected_a in content

    expected_b = param_line_fmt.format('b,]]')
    assert expected_b in content

    expected_c = param_line_fmt.format('c,')
    assert expected_c in content

    expected_d = param_line_fmt.format('d[,')
    assert expected_d in content

    expected_e = param_line_fmt.format('e,')
    assert expected_e in content

    expected_f = param_line_fmt.format('f]')
    assert expected_f in content

    expected_parameter_list_foo = '(\n{}{}{}{}{}{})'.format(
        expected_a,
        expected_b,
        expected_c,
        expected_d,
        expected_e,
        expected_f,
    )
    assert expected_parameter_list_foo in content


@pytest.mark.sphinx('html', testroot='_blank')
def test_module_content_line_number(app):
    text = '.. py:module:: foo\n\n   Some link here: :ref:`abc`\n'
    doc = restructuredtext.parse(app, text)
    xrefs = list(doc.findall(condition=addnodes.pending_xref))
    assert len(xrefs) == 1
    source, line = docutils.utils.get_source_line(xrefs[0])
    assert 'index.rst' in source
    assert line == 3


@pytest.mark.sphinx(
    'html',
    testroot='root',
    freshenv=True,
    confoverrides={'python_display_short_literal_types': True},
)
def test_short_literal_types(app):
    text = """\
.. py:function:: literal_ints(x: Literal[1, 2, 3] = 1) -> None
.. py:function:: literal_union(x: Union[Literal["a"], Literal["b"], Literal["c"]]) -> None
"""
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'literal_ints'],
                            [
                                desc_parameterlist,
                                (
                                    [
                                        desc_parameter,
                                        (
                                            [desc_sig_name, 'x'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [
                                                desc_sig_name,
                                                (
                                                    [desc_sig_literal_number, '1'],
                                                    desc_sig_space,
                                                    [desc_sig_punctuation, '|'],
                                                    desc_sig_space,
                                                    [desc_sig_literal_number, '2'],
                                                    desc_sig_space,
                                                    [desc_sig_punctuation, '|'],
                                                    desc_sig_space,
                                                    [desc_sig_literal_number, '3'],
                                                ),
                                            ],
                                            desc_sig_space,
                                            [desc_sig_operator, '='],
                                            desc_sig_space,
                                            [nodes.inline, '1'],
                                        ),
                                    ],
                                ),
                            ],
                            [desc_returns, pending_xref, 'None'],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'literal_union'],
                            [
                                desc_parameterlist,
                                (
                                    [
                                        desc_parameter,
                                        (
                                            [desc_sig_name, 'x'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [
                                                desc_sig_name,
                                                (
                                                    [desc_sig_literal_string, "'a'"],
                                                    desc_sig_space,
                                                    [desc_sig_punctuation, '|'],
                                                    desc_sig_space,
                                                    [desc_sig_literal_string, "'b'"],
                                                    desc_sig_space,
                                                    [desc_sig_punctuation, '|'],
                                                    desc_sig_space,
                                                    [desc_sig_literal_string, "'c'"],
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            ],
                            [desc_returns, pending_xref, 'None'],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_function_pep_695(app):
    text = """.. py:function:: func[\
        S,\
        T: int,\
        U: (int, str),\
        R: int | str,\
        A: int | Annotated[int, ctype("char")],\
        *V,\
        **P\
    ]
    """
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, 'func'],
                            [
                                desc_type_parameter_list,
                                (
                                    [desc_type_parameter, ([desc_sig_name, 'S'])],
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_name, 'T'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [desc_sig_name, ([pending_xref, 'int'])],
                                        ),
                                    ],
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_name, 'U'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [desc_sig_punctuation, '('],
                                            [
                                                desc_sig_name,
                                                (
                                                    [pending_xref, 'int'],
                                                    [desc_sig_punctuation, ','],
                                                    desc_sig_space,
                                                    [pending_xref, 'str'],
                                                ),
                                            ],
                                            [desc_sig_punctuation, ')'],
                                        ),
                                    ],
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_name, 'R'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [
                                                desc_sig_name,
                                                (
                                                    [pending_xref, 'int'],
                                                    desc_sig_space,
                                                    [desc_sig_punctuation, '|'],
                                                    desc_sig_space,
                                                    [pending_xref, 'str'],
                                                ),
                                            ],
                                        ),
                                    ],
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_name, 'A'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [
                                                desc_sig_name,
                                                (
                                                    [pending_xref, 'int'],
                                                    [desc_sig_space, ' '],
                                                    [desc_sig_punctuation, '|'],
                                                    [desc_sig_space, ' '],
                                                    [pending_xref, 'Annotated'],
                                                    [desc_sig_punctuation, '['],
                                                    [pending_xref, 'int'],
                                                    [desc_sig_punctuation, ','],
                                                    [desc_sig_space, ' '],
                                                    [pending_xref, 'ctype'],
                                                    [desc_sig_punctuation, '('],
                                                    [desc_sig_literal_string, "'char'"],
                                                    [desc_sig_punctuation, ')'],
                                                    [desc_sig_punctuation, ']'],
                                                ),
                                            ],
                                        ),
                                    ],
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_operator, '*'],
                                            [desc_sig_name, 'V'],
                                        ),
                                    ],
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_operator, '**'],
                                            [desc_sig_name, 'P'],
                                        ),
                                    ],
                                ),
                            ],
                            [desc_parameterlist, ()],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_class_def_pep_695(app):
    # Non-concrete unbound generics are allowed at runtime but type checkers
    # should fail (https://peps.python.org/pep-0695/#type-parameter-scopes)
    text = """.. py:class:: Class[S: Sequence[T], T, KT, VT](Dict[KT, VT])"""
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_name, 'Class'],
                            [
                                desc_type_parameter_list,
                                (
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_name, 'S'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [
                                                desc_sig_name,
                                                (
                                                    [pending_xref, 'Sequence'],
                                                    [desc_sig_punctuation, '['],
                                                    [pending_xref, 'T'],
                                                    [desc_sig_punctuation, ']'],
                                                ),
                                            ],
                                        ),
                                    ],
                                    [desc_type_parameter, ([desc_sig_name, 'T'])],
                                    [desc_type_parameter, ([desc_sig_name, 'KT'])],
                                    [desc_type_parameter, ([desc_sig_name, 'VT'])],
                                ),
                            ],
                            [desc_parameterlist, ([desc_parameter, 'Dict[KT, VT]'])],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_class_def_pep_696(app):
    # test default values for type variables without using PEP 696 AST parser
    text = """.. py:class:: Class[\
        T, KT, VT,\
        J: int,\
        K = list,\
        S: str = str,\
        L: (T, tuple[T, ...], collections.abc.Iterable[T]) = set[T],\
        Q: collections.abc.Mapping[KT, VT] = dict[KT, VT],\
        *V = *tuple[*Ts, bool],\
        **P = [int, Annotated[int, ValueRange(3, 10), ctype("char")]]\
    ](Other[T, KT, VT, J, S, L, Q, *V, **P])
    """
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_name, 'Class'],
                            [
                                desc_type_parameter_list,
                                (
                                    [desc_type_parameter, ([desc_sig_name, 'T'])],
                                    [desc_type_parameter, ([desc_sig_name, 'KT'])],
                                    [desc_type_parameter, ([desc_sig_name, 'VT'])],
                                    # J: int
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_name, 'J'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [desc_sig_name, ([pending_xref, 'int'])],
                                        ),
                                    ],
                                    # K = list
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_name, 'K'],
                                            desc_sig_space,
                                            [desc_sig_operator, '='],
                                            desc_sig_space,
                                            [nodes.inline, 'list'],
                                        ),
                                    ],
                                    # S: str = str
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_name, 'S'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [desc_sig_name, ([pending_xref, 'str'])],
                                            desc_sig_space,
                                            [desc_sig_operator, '='],
                                            desc_sig_space,
                                            [nodes.inline, 'str'],
                                        ),
                                    ],
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_name, 'L'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [desc_sig_punctuation, '('],
                                            [
                                                desc_sig_name,
                                                (
                                                    # T
                                                    [pending_xref, 'T'],
                                                    [desc_sig_punctuation, ','],
                                                    desc_sig_space,
                                                    # tuple[T, ...]
                                                    [pending_xref, 'tuple'],
                                                    [desc_sig_punctuation, '['],
                                                    [pending_xref, 'T'],
                                                    [desc_sig_punctuation, ','],
                                                    desc_sig_space,
                                                    [desc_sig_punctuation, '...'],
                                                    [desc_sig_punctuation, ']'],
                                                    [desc_sig_punctuation, ','],
                                                    desc_sig_space,
                                                    # collections.abc.Iterable[T]
                                                    [
                                                        pending_xref,
                                                        'collections.abc.Iterable',
                                                    ],
                                                    [desc_sig_punctuation, '['],
                                                    [pending_xref, 'T'],
                                                    [desc_sig_punctuation, ']'],
                                                ),
                                            ],
                                            [desc_sig_punctuation, ')'],
                                            desc_sig_space,
                                            [desc_sig_operator, '='],
                                            desc_sig_space,
                                            [nodes.inline, 'set[T]'],
                                        ),
                                    ],
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_name, 'Q'],
                                            [desc_sig_punctuation, ':'],
                                            desc_sig_space,
                                            [
                                                desc_sig_name,
                                                (
                                                    [
                                                        pending_xref,
                                                        'collections.abc.Mapping',
                                                    ],
                                                    [desc_sig_punctuation, '['],
                                                    [pending_xref, 'KT'],
                                                    [desc_sig_punctuation, ','],
                                                    desc_sig_space,
                                                    [pending_xref, 'VT'],
                                                    [desc_sig_punctuation, ']'],
                                                ),
                                            ],
                                            desc_sig_space,
                                            [desc_sig_operator, '='],
                                            desc_sig_space,
                                            [nodes.inline, 'dict[KT, VT]'],
                                        ),
                                    ],
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_operator, '*'],
                                            [desc_sig_name, 'V'],
                                            desc_sig_space,
                                            [desc_sig_operator, '='],
                                            desc_sig_space,
                                            [nodes.inline, '*tuple[*Ts, bool]'],
                                        ),
                                    ],
                                    [
                                        desc_type_parameter,
                                        (
                                            [desc_sig_operator, '**'],
                                            [desc_sig_name, 'P'],
                                            desc_sig_space,
                                            [desc_sig_operator, '='],
                                            desc_sig_space,
                                            [
                                                nodes.inline,
                                                '[int, Annotated[int, ValueRange(3, 10), ctype("char")]]',
                                            ],
                                        ),
                                    ],
                                ),
                            ],
                            [
                                desc_parameterlist,
                                (
                                    [
                                        desc_parameter,
                                        'Other[T, KT, VT, J, S, L, Q, *V, **P]',
                                    ],
                                ),
                            ],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )


@pytest.mark.parametrize(
    ('tp_list', 'tptext'),
    [
        ('[T:int]', '[T: int]'),
        ('[T:*Ts]', '[T: *Ts]'),
        ('[T:int|(*Ts)]', '[T: int | (*Ts)]'),
        ('[T:(*Ts)|int]', '[T: (*Ts) | int]'),
        ('[T:(int|(*Ts))]', '[T: (int | (*Ts))]'),
        ('[T:((*Ts)|int)]', '[T: ((*Ts) | int)]'),
        ("[T:Annotated[int,ctype('char')]]", "[T: Annotated[int, ctype('char')]]"),
    ],
)
@pytest.mark.sphinx('html', testroot='root')
def test_pep_695_and_pep_696_whitespaces_in_bound(app, tp_list, tptext):
    text = f'.. py:function:: f{tp_list}()'
    doctree = restructuredtext.parse(app, text)
    assert doctree.astext() == f'\n\nf{tptext}()\n\n'

    text = f'.. py:function:: f{tp_list}() -> Annotated[T, Qux[int]()]'
    doctree = restructuredtext.parse(app, text)
    assert doctree.astext() == f'\n\nf{tptext}() -> Annotated[T, Qux[int]()]\n\n'


@pytest.mark.parametrize(
    ('tp_list', 'tptext'),
    [
        ('[T:(int,str)]', '[T: (int, str)]'),
        ('[T:(int|str,*Ts)]', '[T: (int | str, *Ts)]'),
    ],
)
@pytest.mark.sphinx('html', testroot='root')
def test_pep_695_and_pep_696_whitespaces_in_constraints(app, tp_list, tptext):
    text = f'.. py:function:: f{tp_list}()'
    doctree = restructuredtext.parse(app, text)
    assert doctree.astext() == f'\n\nf{tptext}()\n\n'

    text = f'.. py:function:: f{tp_list}() -> Annotated[T, Qux[int]()]'
    doctree = restructuredtext.parse(app, text)
    assert doctree.astext() == f'\n\nf{tptext}() -> Annotated[T, Qux[int]()]\n\n'


@pytest.mark.parametrize(
    ('tp_list', 'tptext'),
    [
        ('[T=int]', '[T = int]'),
        ('[T:int=int]', '[T: int = int]'),
        ('[*V=*Ts]', '[*V = *Ts]'),
        ('[*V=(*Ts)]', '[*V = (*Ts)]'),
        ('[*V=*tuple[str,...]]', '[*V = *tuple[str, ...]]'),
        ('[*V=*tuple[*Ts,...]]', '[*V = *tuple[*Ts, ...]]'),
        ('[*V=*tuple[int,*Ts]]', '[*V = *tuple[int, *Ts]]'),
        ('[*V=*tuple[*Ts,int]]', '[*V = *tuple[*Ts, int]]'),
        ('[**P=[int,*Ts]]', '[**P = [int, *Ts]]'),
        ('[**P=[int, int*3]]', '[**P = [int, int * 3]]'),
        ('[**P=[int, *Ts*3]]', '[**P = [int, *Ts * 3]]'),
        ('[**P=[int,A[int,ctype("char")]]]', '[**P = [int, A[int, ctype("char")]]]'),
    ],
)
@pytest.mark.sphinx('html', testroot='root')
def test_pep_695_and_pep_696_whitespaces_in_default(app, tp_list, tptext):
    text = f'.. py:function:: f{tp_list}()'
    doctree = restructuredtext.parse(app, text)
    assert doctree.astext() == f'\n\nf{tptext}()\n\n'

    text = f'.. py:function:: f{tp_list}() -> Annotated[T, Qux[int]()]'
    doctree = restructuredtext.parse(app, text)
    assert doctree.astext() == f'\n\nf{tptext}() -> Annotated[T, Qux[int]()]\n\n'



#!/usr/bin/env python3

from __future__ import annotations

import argparse
import dataclasses
import re
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from typing import Literal


script_dir = Path(__file__).resolve().parent
package_dir = script_dir.parent


@dataclasses.dataclass(frozen=True, slots=True)
class VersionInfo:
    major: int
    minor: int
    micro: int
    level: Literal['a', 'b', 'rc', 'final']
    serial: int

    @property
    def releaselevel(self) -> Literal['alpha', 'beta', 'candidate', 'final']:
        if self.level == 'final':
            return 'final'
        if self.level == 'a':
            return 'alpha'
        if self.level == 'b':
            return 'beta'
        if self.level == 'rc':
            return 'candidate'
        msg = f'Unknown release level: {self.level}'
        raise RuntimeError(msg)

    @property
    def is_final(self) -> bool:
        return self.level == 'final'

    @property
    def version(self) -> str:
        return f'{self.major}.{self.minor}.{self.micro}'

    @property
    def release(self) -> str:
        return f'{self.major}.{self.minor}.{self.micro}{self.level}{self.serial}'

    @property
    def version_tuple(self) -> tuple[int, int, int]:
        return self.major, self.minor, self.micro

    @property
    def release_tuple(self) -> tuple[int, int, int, str, int]:
        return self.major, self.minor, self.micro, self.releaselevel, self.serial


def parse_version(version: str) -> VersionInfo:
    # Final version:
    # - "X.Y.Z" -> (X, Y, Z, 'final', 0)
    # - "X.Y" -> (X, Y, 0, 'final', 0) [shortcut]
    if matched := re.fullmatch(r'(\d+)\.(\d+)(?:\.(\d+))?', version):
        major, minor, micro = matched.groups(default='0')
        return VersionInfo(int(major), int(minor), int(micro), 'final', 0)

    # Pre-release versions:
    # - "X.Y.ZaN" -> (X, Y, Z, 'alpha', N)
    # - "X.Y.ZbN" -> (X, Y, Z, 'beta', N)
    # - "X.Y.ZrcN" -> (X, Y, Z, 'candidate', N)
    if matched := re.fullmatch(r'(\d+)\.(\d+)\.(\d+)(a|b|rc)(\d+)', version):
        major, minor, micro, level, serial = matched.groups()
        return VersionInfo(int(major), int(minor), int(micro), level, int(serial))  # type: ignore[arg-type]

    msg = f'Unknown version: {version}'
    raise RuntimeError(msg)


def bump_version(
    path: Path, version_info: VersionInfo, in_develop: bool = True
) -> None:
    if in_develop or version_info.is_final:
        version = version_info.version
    else:
        version = version_info.release

    with open(path, encoding='utf-8') as f:
        lines = f.read().splitlines(keepends=True)

    for i, line in enumerate(lines):
        if line.startswith('__version__: Final = '):
            lines[i] = f"__version__: Final = '{version}'\n"
            continue
        if line.startswith('version_info: Final = '):
            lines[i] = f'version_info: Final = {version_info.release_tuple}\n'
            continue
        if line.startswith('_in_development = '):
            lines[i] = f'_in_development = {in_develop}\n'
            continue

    with open(path, 'w', encoding='utf-8') as f:
        f.writelines(lines)


class Changes:
    def __init__(self, path: Path) -> None:
        self.path = path

        with open(self.path, encoding='utf-8') as f:
            version = f.readline().strip()
            matched = re.fullmatch(r'Release (.*) \((.*)\)', version)
            if matched is None:
                msg = f'Unknown CHANGES format: {version}'
                raise RuntimeError(msg)

            self.version, release_date = matched.groups()
            self.in_development = release_date == 'in development'
            self.version_tuple = parse_version(self.version).version_tuple

    def finalise_release_date(self) -> None:
        release_date = time.strftime('%b %d, %Y')
        heading = f'Release {self.version} (released {release_date})'
        with open(self.path, 'r+', encoding='utf-8') as f:
            f.readline()  # skip first two lines
            f.readline()
            body = f.read()

            f.seek(0)
            f.truncate(0)
            f.write(heading + '\n')
            f.write('=' * len(heading) + '\n')
            f.write(self.filter_empty_sections(body))

    def add_release(self, version_info: VersionInfo) -> None:
        heading = f'Release {version_info.version} (in development)'
        tmpl = (script_dir / 'CHANGES_template.rst').read_text(encoding='utf-8')
        with open(self.path, 'r+', encoding='utf-8') as f:
            body = f.read()

            f.seek(0)
            f.truncate(0)
            f.write(heading + '\n')
            f.write('=' * len(heading) + '\n')
            f.write('\n')
            f.write(tmpl)
            f.write('\n')
            f.write(body)

    @staticmethod
    def filter_empty_sections(body: str) -> str:
        return re.sub(
            '^\n.+\n-{3,}\n+(?=\n.+\n[-=]{3,}\n)', '', body, flags=re.MULTILINE
        )


class Skip(Exception):
    pass


@contextmanager
def processing(message: str) -> Iterator[None]:
    try:
        print(message + ' ... ', end='')
        yield
    except Skip as exc:
        print(f'skip: {exc}')
    except Exception:
        print('error')
        raise
    else:
        print('done')


def parse_options(argv: Sequence[str]) -> tuple[VersionInfo, bool]:
    parser = argparse.ArgumentParser()
    parser.add_argument('version', help='A version number (cf. 1.6.0b0)')
    parser.add_argument('--in-develop', action='store_true')
    options = parser.parse_args(argv)
    return parse_version(options.version), options.in_develop


def main() -> None:
    version, in_develop = parse_options(sys.argv[1:])

    with processing('Rewriting sphinx/__init__.py'):
        bump_version(package_dir / 'sphinx' / '__init__.py', version, in_develop)

    with processing('Rewriting CHANGES'):
        changes = Changes(package_dir / 'CHANGES.rst')
        if changes.version_tuple == version.version_tuple:
            if changes.in_development and version.is_final and not in_develop:
                changes.finalise_release_date()
            else:
                reason = 'version not changed'
                raise Skip(reason)
        else:
            if changes.in_development:
                print(f'WARNING: last version is not released yet: {changes.version}')
            changes.add_release(version)


if __name__ == '__main__':
    main()



"""Test object description directives."""

from __future__ import annotations

from typing import TYPE_CHECKING

import docutils.utils
import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.io import create_publisher
from sphinx.testing import restructuredtext
from sphinx.util.docutils import sphinx_domains

if TYPE_CHECKING:
    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment
    from sphinx.testing.util import SphinxTestApp


def _doctree_for_test(
    app: Sphinx, env: BuildEnvironment, docname: str
) -> nodes.document:
    env.prepare_settings(docname)
    publisher = create_publisher(app, 'restructuredtext')
    with sphinx_domains(env):
        publisher.set_source(source_path=str(env.doc2path(docname)))
        publisher.publish()
        return publisher.document


@pytest.mark.sphinx('text', testroot='object-description-sections')
def test_object_description_sections(app: SphinxTestApp) -> None:
    doctree = _doctree_for_test(app, app.env, 'index')
    # <document>
    #     <index>
    #     <desc>
    #         <desc_signature>
    #             <desc_name>
    #                 func
    #             <desc_parameterlist>
    #         <desc_content>
    #             <section>
    #                 <title>
    #                     Overview
    #                 <paragraph>
    #                     Lorem ipsum dolar sit amet

    assert isinstance(doctree[0], addnodes.index)
    assert isinstance(doctree[1], addnodes.desc)
    assert isinstance(doctree[1][0], addnodes.desc_signature)
    assert isinstance(doctree[1][1], addnodes.desc_content)
    assert isinstance(doctree[1][1][0], nodes.section)
    assert isinstance(doctree[1][1][0][0], nodes.title)
    assert doctree[1][1][0][0][0] == 'Overview'
    assert isinstance(doctree[1][1][0][1], nodes.paragraph)
    assert doctree[1][1][0][1][0] == 'Lorem ipsum dolar sit amet'


@pytest.mark.sphinx('html', testroot='_blank')
def test_object_description_content_line_number(app: SphinxTestApp) -> None:
    text = '.. py:function:: foo(bar)\n\n   Some link here: :ref:`abc`\n'
    doc = restructuredtext.parse(app, text)
    xrefs = list(doc.findall(condition=addnodes.pending_xref))
    assert len(xrefs) == 1
    source, line = docutils.utils.get_source_line(xrefs[0])
    assert 'index.rst' in source
    assert line == 3



from __future__ import annotations

from typing import TYPE_CHECKING

import pygments
import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx(
    'html',
    testroot='reST-code-block',
    confoverrides={'html_codeblock_linenos_style': 'table'},
)
def test_html_codeblock_linenos_style_table(app: SphinxTestApp) -> None:
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert (
        '<div class="linenodiv"><pre><span class="normal">1</span>\n'
        '<span class="normal">2</span>\n'
        '<span class="normal">3</span>\n'
        '<span class="normal">4</span></pre></div>'
    ) in content


@pytest.mark.sphinx(
    'html',
    testroot='reST-code-block',
    confoverrides={'html_codeblock_linenos_style': 'inline'},
)
def test_html_codeblock_linenos_style_inline(app: SphinxTestApp) -> None:
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    assert '<span class="linenos">1</span>' in content


@pytest.mark.sphinx('html', testroot='reST-code-role')
def test_html_code_role(app: SphinxTestApp) -> None:
    if tuple(map(int, pygments.__version__.split('.')[:2])) >= (2, 19):
        sp = '<span class="w"> </span>'
    else:
        sp = ' '

    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')

    common_content = (
        f'<span class="k">def</span>{sp}<span class="nf">foo</span>'
        '<span class="p">(</span>'
        '<span class="mi">1</span> '
        '<span class="o">+</span> '
        '<span class="mi">2</span> '
        '<span class="o">+</span> '
        '<span class="kc">None</span> '
        '<span class="o">+</span> '
        '<span class="s2">&quot;abc&quot;</span>'
        '<span class="p">):</span> '
        '<span class="k">pass</span>'
    )
    assert (
        '<p>Inline <code class="code highlight python docutils literal highlight-python">'
        + common_content
        + '</code> code block</p>'
    ) in content
    assert (
        '<div class="highlight-python notranslate">'
        '<div class="highlight"><pre><span></span>' + common_content
    ) in content



from __future__ import annotations

import gettext
import os
import sys
from pathlib import Path
from types import SimpleNamespace
from typing import TYPE_CHECKING

import docutils
import pytest

import sphinx
import sphinx.locale
import sphinx.pycode
from sphinx.testing.util import _clean_up_global_state

if TYPE_CHECKING:
    from collections.abc import Iterator

_TESTS_ROOT = Path(__file__).resolve().parent
if 'CI' in os.environ and (_TESTS_ROOT / 'roots-read-only').is_dir():
    _ROOTS_DIR = _TESTS_ROOT / 'roots-read-only'
else:
    _ROOTS_DIR = _TESTS_ROOT / 'roots'


def _init_console(
    locale_dir: str | os.PathLike[str] | None = sphinx.locale._LOCALE_DIR,
    catalog: str = 'sphinx',
) -> tuple[gettext.NullTranslations, bool]:
    """Monkeypatch ``init_console`` to skip its action.

    Some tests rely on warning messages in English. We don't want
    CLI tests to bleed over those tests and make their warnings
    translated.
    """
    return gettext.NullTranslations(), False


sphinx.locale.init_console = _init_console

pytest_plugins = ['sphinx.testing.fixtures']

# Exclude 'roots' dirs for pytest test collector
collect_ignore = ['roots']

os.environ['SPHINX_AUTODOC_RELOAD_MODULES'] = '1'


@pytest.fixture(scope='session')
def rootdir() -> Path:
    return _ROOTS_DIR


def pytest_report_header(config: pytest.Config) -> str:
    lines = [
        f'libraries: Sphinx-{sphinx.__display_version__}, docutils-{docutils.__version__}',
    ]
    if sys.version_info[:2] >= (3, 13):
        lines.append(f'GIL enabled?: {sys._is_gil_enabled()}')
    lines.append(f'test roots directory: {_ROOTS_DIR}')
    if hasattr(config, '_tmp_path_factory'):
        lines.append(f'base tmp_path: {config._tmp_path_factory.getbasetemp()}')
    return '\n'.join(lines)


@pytest.fixture(autouse=True)
def _cleanup_docutils() -> Iterator[None]:
    saved_path = sys.path
    yield  # run the test
    sys.path[:] = saved_path

    _clean_up_global_state()


@pytest.fixture
def _http_teapot(monkeypatch: pytest.MonkeyPatch) -> Iterator[None]:
    """Short-circuit HTTP requests.

    Windows takes too long to fail on connections, hence this fixture.
    """
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/418
    response = SimpleNamespace(status_code=418)

    def _request(*args, **kwargs):
        return response

    with monkeypatch.context() as m:
        m.setattr('sphinx.util.requests._Session.request', _request)
        yield


@pytest.fixture
def make_app_with_empty_project(make_app, tmp_path):
    (tmp_path / 'conf.py').touch()

    def _make_app(*args, **kw):
        kw.setdefault('srcdir', Path(tmp_path))
        return make_app(*args, **kw)

    return _make_app



"""Test the patched directives."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from docutils import nodes

from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='_blank')
def test_code_directive(app: SphinxTestApp) -> None:
    # normal case
    text = '.. code::\n\n   print("hello world")\n'

    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, [nodes.document, nodes.literal_block, 'print("hello world")'])
    assert_node(doctree[0], language='default', highlight_args={})

    # with language
    text = '.. code:: python\n\n   print("hello world")\n'

    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, [nodes.document, nodes.literal_block, 'print("hello world")'])
    assert_node(doctree[0], language='python', highlight_args={})

    # :number-lines: option
    text = '.. code:: python\n   :number-lines:\n\n   print("hello world")\n'

    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, [nodes.document, nodes.literal_block, 'print("hello world")'])
    assert_node(doctree[0], language='python', linenos=True, highlight_args={})

    # :number-lines: option
    text = '.. code:: python\n   :number-lines: 5\n\n   print("hello world")\n'

    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, [nodes.document, nodes.literal_block, 'print("hello world")'])
    assert_node(
        doctree[0], language='python', linenos=True, highlight_args={'linenostart': 5}
    )


@pytest.mark.sphinx('html', testroot='directive-csv-table')
def test_csv_table_directive(app):
    # relative path from current document
    text = '.. csv-table::\n   :file: example.csv\n'
    doctree = restructuredtext.parse(app, text, docname='subdir/index')
    assert_node(
        doctree,
        (
            [
                nodes.table,
                nodes.tgroup,
                (nodes.colspec, nodes.colspec, nodes.colspec, [nodes.tbody, nodes.row]),
            ],
        ),
    )
    assert_node(
        doctree[0][0][3][0],
        (
            [nodes.entry, nodes.paragraph, 'FOO'],
            [nodes.entry, nodes.paragraph, 'BAR'],
            [nodes.entry, nodes.paragraph, 'BAZ'],
        ),
    )

    # absolute path from source directory
    text = '.. csv-table::\n   :file: /example.csv\n'
    doctree = restructuredtext.parse(app, text, docname='subdir/index')
    assert_node(
        doctree,
        (
            [
                nodes.table,
                nodes.tgroup,
                (nodes.colspec, nodes.colspec, nodes.colspec, [nodes.tbody, nodes.row]),
            ],
        ),
    )
    assert_node(
        doctree[0][0][3][0],
        (
            [nodes.entry, nodes.paragraph, 'foo'],
            [nodes.entry, nodes.paragraph, 'bar'],
            [nodes.entry, nodes.paragraph, 'baz'],
        ),
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_math_directive(app: SphinxTestApp) -> None:
    # normal case
    text = '.. math:: E = mc^2'
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, [nodes.document, nodes.math_block, 'E = mc^2\n\n'])

    # :name: option
    text = '.. math:: E = mc^2\n   :name: eq1\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree, [nodes.document, (nodes.target, [nodes.math_block, 'E = mc^2\n\n'])]
    )
    assert_node(doctree[1], nodes.math_block, docname='index', label='eq1', number=1)

    # :label: option
    text = '.. math:: E = mc^2\n   :label: eq2\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree, [nodes.document, (nodes.target, [nodes.math_block, 'E = mc^2\n\n'])]
    )
    assert_node(doctree[1], nodes.math_block, docname='index', label='eq2', number=2)

    # :label: option without value
    text = '.. math:: E = mc^2\n   :label:\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree, [nodes.document, (nodes.target, [nodes.math_block, 'E = mc^2\n\n'])]
    )
    assert_node(
        doctree[1],
        nodes.math_block,
        ids=['equation-index-0'],
        docname='index',
        label='index:0',
        number=3,
    )



from __future__ import annotations

import itertools
import operator
from typing import TYPE_CHECKING

from sphinx._cli.util.colour import blue, reset
from sphinx._cli.util.errors import strip_escape_sequences

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Final

CURSOR_UP: Final[str] = '\x1b[2A'  # ignored ANSI code
ERASE_LINE: Final[str] = '\x1b[2K'  # supported ANSI code
TEXT: Final[str] = '\x07 ß Hello world!'


def test_strip_escape_sequences() -> None:
    # double ERASE_LINE so that the tested strings may have 2 of them
    ansi_base_blocks = [
        TEXT,
        blue(TEXT),
        reset(TEXT),
        ERASE_LINE,
        ERASE_LINE,
        CURSOR_UP,
    ]
    # :func:`strip_escape_sequences` strips ANSI codes known by Sphinx
    text_base_blocks = [
        TEXT,
        TEXT,
        TEXT,
        '',
        '',
        CURSOR_UP,
    ]

    assert len(text_base_blocks) == len(ansi_base_blocks)
    N = len(ansi_base_blocks)

    def next_ansi_blocks(choices: Sequence[str], n: int) -> Sequence[str]:
        # Get a list of *n* words from a cyclic sequence of *choices*.
        #
        # For instance ``next_ansi_blocks(['a', 'b'], 3) == ['a', 'b', 'a']``.
        stream = itertools.cycle(choices)
        return list(map(operator.itemgetter(0), zip(stream, range(n), strict=False)))

    # generate all permutations of length N
    for sigma in itertools.permutations(range(N), N):
        # apply the permutation on the blocks with ANSI codes
        ansi_blocks = list(map(ansi_base_blocks.__getitem__, sigma))
        # apply the permutation on the blocks with stripped codes
        text_blocks = list(map(text_base_blocks.__getitem__, sigma))

        for glue, n in itertools.product(['.', '\n', '\r\n'], range(4 * N)):
            ansi_strings = next_ansi_blocks(ansi_blocks, n)
            text_strings = next_ansi_blocks(text_blocks, n)
            assert len(ansi_strings) == len(text_strings) == n

            ansi_string = glue.join(ansi_strings)
            text_string = glue.join(text_strings)
            assert strip_escape_sequences(ansi_string) == text_string


def test_strip_ansi_short_forms() -> None:
    # In Sphinx, we always "normalize" the color codes so that they
    # match "\x1b\[(\d\d;){0,2}(\d\d)m" but it might happen that
    # some messages use '\x1b[0m' instead of ``reset(s)``, so we
    # test whether this alternative form is supported or not.

    for strip_function in strip_escape_sequences, strip_escape_sequences:
        # \x1b[m and \x1b[0m are equivalent to \x1b[00m
        assert strip_function('\x1b[m') == ''
        assert strip_function('\x1b[0m') == ''

        # \x1b[1m is equivalent to \x1b[01m
        assert strip_function('\x1b[1mbold\x1b[0m') == 'bold'

    # \x1b[K is equivalent to \x1b[0K
    assert strip_escape_sequences('\x1b[K') == ''



#!/usr/bin/env python3
from __future__ import annotations

import shutil
import subprocess
from pathlib import Path

SPHINX_ROOT = Path(__file__).resolve().parent.parent
TEST_JS_FIXTURES = SPHINX_ROOT / 'tests' / 'js' / 'fixtures'
TEST_JS_ROOTS = [
    directory
    for directory in (SPHINX_ROOT / 'tests' / 'js' / 'roots').iterdir()
    if (directory / 'conf.py').exists()
]


def build(srcdir: Path) -> None:
    cmd = (
        'sphinx-build',
        '--fresh-env',
        '--quiet',
        *('--builder', 'html'),
        f'{srcdir}',
        f'{srcdir}/_build',
    )
    subprocess.run(cmd, check=True, capture_output=True)


for directory in TEST_JS_ROOTS:
    searchindex = directory / '_build' / 'searchindex.js'
    destination = TEST_JS_FIXTURES / directory.name / 'searchindex.js'

    print(f'Building {directory} ... ', end='')
    build(directory)
    print('done')

    print(f'Copying {searchindex} to {destination} ... ', end='')
    destination.parent.mkdir(exist_ok=True, parents=True)
    shutil.copy2(searchindex, destination)
    print('done')



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={'option_emphasise_placeholders': True},
)
def test_option_emphasise_placeholders(app: SphinxTestApp) -> None:
    app.build()
    content = (app.outdir / 'objects.html').read_text(encoding='utf8')
    assert '<em><span class="pre">TYPE</span></em>' in content
    assert '{TYPE}' not in content
    assert (
        '<em><span class="pre">WHERE</span></em>'
        '<span class="pre">-</span>'
        '<em><span class="pre">COUNT</span></em>'
    ) in content
    assert '<span class="pre">{{value}}</span>' in content
    assert (
        '<span class="pre">--plugin.option</span></span>'
        '<a class="headerlink" href="#cmdoption-perl-plugin.option" title="Link to this definition">¶</a></dt>'
    ) in content


@pytest.mark.sphinx('html', testroot='root')
def test_option_emphasise_placeholders_default(app: SphinxTestApp) -> None:
    app.build()
    content = (app.outdir / 'objects.html').read_text(encoding='utf8')
    assert '<span class="pre">={TYPE}</span>' in content
    assert '<span class="pre">={WHERE}-{COUNT}</span></span>' in content
    assert '<span class="pre">{client_name}</span>' in content
    assert (
        '<span class="pre">--plugin.option</span></span>'
        '<span class="sig-prename descclassname"></span>'
        '<a class="headerlink" href="#cmdoption-perl-plugin.option" title="Link to this definition">¶</a></dt>'
    ) in content


@pytest.mark.sphinx('html', testroot='root')
def test_option_reference_with_value(app: SphinxTestApp) -> None:
    app.build()
    content = (app.outdir / 'objects.html').read_text(encoding='utf-8')
    assert (
        '<span class="pre">-mapi</span></span><span class="sig-prename descclassname">'
        '</span><a class="headerlink" href="#cmdoption-git-commit-mapi"'
    ) in content
    assert (
        'first option <a class="reference internal" href="#cmdoption-git-commit-mapi">'
    ) in content
    assert (
        '<a class="reference internal" href="#cmdoption-git-commit-mapi">'
        '<code class="xref std std-option docutils literal notranslate"><span class="pre">-mapi[=xxx]</span></code></a>'
    ) in content
    assert (
        '<span class="pre">-mapi</span> <span class="pre">with_space</span>'
    ) in content



"""Test all builders.

This test skips building docs for some builders that have independent testcases.
(html, changes, epub, latex, texinfo and manpage)
"""

from __future__ import annotations

import shutil
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from sphinx.testing.util import SphinxTestApp

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path
    from unittest.mock import MagicMock

    from sphinx.testing.util import SphinxTestApp


def request_session_head(url: str, **kwargs: object) -> mock.Mock:
    response = mock.Mock()
    response.status_code = 200
    response.url = url
    return response


@pytest.fixture
def nonascii_srcdir(rootdir: Path, sphinx_test_tempdir: Path) -> Path:
    # Build in a non-ASCII source dir
    test_name = '\u65e5\u672c\u8a9e'
    srcdir = sphinx_test_tempdir / 'test_build_all' / test_name
    if not srcdir.exists():
        shutil.copytree(rootdir / 'test-root', srcdir)

    # add a doc with a non-ASCII file name to the source dir
    (srcdir / f'{test_name}.txt').write_text(
        'non-ascii file name page\n========================\n', encoding='utf8'
    )

    with srcdir.joinpath('index.txt').open('a', encoding='utf8') as f:
        f.write(f"""
.. toctree::

   {test_name}/{test_name}
""")
    return srcdir


def test_build_dirhtml(
    make_app: Callable[..., SphinxTestApp], nonascii_srcdir: Path
) -> None:
    app = make_app('dirhtml', srcdir=nonascii_srcdir)
    app.build(force_all=True)


def test_build_singlehtml(
    make_app: Callable[..., SphinxTestApp], nonascii_srcdir: Path
) -> None:
    app = make_app('singlehtml', srcdir=nonascii_srcdir)
    app.build(force_all=True)


def test_build_text(
    make_app: Callable[..., SphinxTestApp], nonascii_srcdir: Path
) -> None:
    app = make_app('text', srcdir=nonascii_srcdir)
    app.build(force_all=True)


def test_build_xml(
    make_app: Callable[..., SphinxTestApp], nonascii_srcdir: Path
) -> None:
    app = make_app('xml', srcdir=nonascii_srcdir)
    app.build(force_all=True)


def test_build_pseudoxml(
    make_app: Callable[..., SphinxTestApp], nonascii_srcdir: Path
) -> None:
    app = make_app('pseudoxml', srcdir=nonascii_srcdir)
    app.build(force_all=True)


@mock.patch(
    'sphinx.builders.linkcheck.requests.head',
    side_effect=request_session_head,
)
def test_build_linkcheck(
    requests_head: MagicMock,
    make_app: Callable[..., SphinxTestApp],
    nonascii_srcdir: Path,
) -> None:
    app = make_app('linkcheck', srcdir=nonascii_srcdir)
    app.build(force_all=True)



"""Test logging util."""

from __future__ import annotations

import codecs
import os
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from docutils import nodes

from sphinx._cli.util.errors import strip_escape_sequences
from sphinx.cmd.build import build_main
from sphinx.util import logging
from sphinx.util.console import colorize
from sphinx.util.logging import is_suppressed_warning, prefixed_warnings
from sphinx.util.parallel import ParallelTasks

from tests.utils import TESTS_ROOT

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('html', testroot='root')
def test_info_and_warning(app: SphinxTestApp) -> None:
    app.verbosity = 2
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.debug('message1')
    logger.info('message2')
    logger.warning('message3')
    logger.critical('message4')
    logger.error('message5')

    assert 'message1' in app.status.getvalue()
    assert 'message2' in app.status.getvalue()
    assert 'message3' not in app.status.getvalue()
    assert 'message4' not in app.status.getvalue()
    assert 'message5' not in app.status.getvalue()

    assert 'message1' not in app.warning.getvalue()
    assert 'message2' not in app.warning.getvalue()
    assert 'WARNING: message3' in app.warning.getvalue()
    assert 'CRITICAL: message4' in app.warning.getvalue()
    assert 'ERROR: message5' in app.warning.getvalue()


@pytest.mark.sphinx('html', testroot='root')
def test_Exception(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.info(Exception)
    assert "<class 'Exception'>" in app.status.getvalue()


@pytest.mark.sphinx('html', testroot='root')
def test_verbosity_filter(app: SphinxTestApp) -> None:
    # verbosity = 0: INFO
    app.verbosity = 0
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.info('message1')
    logger.verbose('message2')
    logger.debug('message3')

    assert 'message1' in app.status.getvalue()
    assert 'message2' not in app.status.getvalue()
    assert 'message3' not in app.status.getvalue()
    assert 'message4' not in app.status.getvalue()

    # verbosity = 1: VERBOSE
    app.verbosity = 1
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.info('message1')
    logger.verbose('message2')
    logger.debug('message3')

    assert 'message1' in app.status.getvalue()
    assert 'message2' in app.status.getvalue()
    assert 'message3' not in app.status.getvalue()
    assert 'message4' not in app.status.getvalue()

    # verbosity = 2: DEBUG
    app.verbosity = 2
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.info('message1')
    logger.verbose('message2')
    logger.debug('message3')

    assert 'message1' in app.status.getvalue()
    assert 'message2' in app.status.getvalue()
    assert 'message3' in app.status.getvalue()
    assert 'message4' not in app.status.getvalue()


@pytest.mark.sphinx('html', testroot='root')
def test_nonl_info_log(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.info('message1', nonl=True)
    logger.info('message2')
    logger.info('message3')

    assert 'message1message2\nmessage3' in app.status.getvalue()


@pytest.mark.sphinx('html', testroot='root')
def test_once_warning_log(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.warning('message: %d', 1, once=True)
    logger.warning('message: %d', 1, once=True)
    logger.warning('message: %d', 2, once=True)

    warnings = strip_escape_sequences(app.warning.getvalue())
    assert 'WARNING: message: 1\nWARNING: message: 2\n' in warnings


def test_is_suppressed_warning():
    suppress_warnings = ['ref', 'files.*', 'rest.duplicated_labels']

    assert is_suppressed_warning(None, None, suppress_warnings) is False
    assert is_suppressed_warning('ref', None, suppress_warnings) is True
    assert is_suppressed_warning('ref', 'numref', suppress_warnings) is True
    assert is_suppressed_warning('ref', 'option', suppress_warnings) is True
    assert is_suppressed_warning('files', 'image', suppress_warnings) is True
    assert is_suppressed_warning('files', 'stylesheet', suppress_warnings) is True
    assert is_suppressed_warning('rest', None, suppress_warnings) is False
    assert is_suppressed_warning('rest', 'syntax', suppress_warnings) is False
    assert is_suppressed_warning('rest', 'duplicated_labels', suppress_warnings) is True


@pytest.mark.sphinx('html', testroot='root')
def test_suppress_warnings(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    app._warncount = 0  # force reset

    app.config.suppress_warnings = []
    app.warning.truncate(0)
    logger.warning('message0', type='test')
    logger.warning('message1', type='test', subtype='logging')
    logger.warning('message2', type='test', subtype='crash')
    logger.warning('message3', type='actual', subtype='logging')
    assert 'message0' in app.warning.getvalue()
    assert 'message1' in app.warning.getvalue()
    assert 'message2' in app.warning.getvalue()
    assert 'message3' in app.warning.getvalue()
    assert app._warncount == 4

    app.config.suppress_warnings = ['test']
    app.warning.truncate(0)
    logger.warning('message0', type='test')
    logger.warning('message1', type='test', subtype='logging')
    logger.warning('message2', type='test', subtype='crash')
    logger.warning('message3', type='actual', subtype='logging')
    assert 'message0' not in app.warning.getvalue()
    assert 'message1' not in app.warning.getvalue()
    assert 'message2' not in app.warning.getvalue()
    assert 'message3' in app.warning.getvalue()
    assert app._warncount == 5

    app.config.suppress_warnings = ['test.logging']
    app.warning.truncate(0)
    logger.warning('message0', type='test')
    logger.warning('message1', type='test', subtype='logging')
    logger.warning('message2', type='test', subtype='crash')
    logger.warning('message3', type='actual', subtype='logging')
    assert 'message0' in app.warning.getvalue()
    assert 'message1' not in app.warning.getvalue()
    assert 'message2' in app.warning.getvalue()
    assert 'message3' in app.warning.getvalue()
    assert app._warncount == 8


@pytest.mark.sphinx('html', testroot='root')
def test_info_location(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.info('message1', location='index')
    assert 'index.txt: message1' in app.status.getvalue()

    logger.info('message2', location=('index', 10))
    assert 'index.txt:10: message2' in app.status.getvalue()

    logger.info('message3', location=None)
    assert '\nmessage3' in app.status.getvalue()

    node = nodes.Element()
    node.source, node.line = ('index.txt', 10)
    logger.info('message4', location=node)
    assert 'index.txt:10: message4' in app.status.getvalue()

    node.source, node.line = ('index.txt', None)
    logger.info('message5', location=node)
    assert 'index.txt:: message5' in app.status.getvalue()

    node.source, node.line = (None, 10)
    logger.info('message6', location=node)
    assert '<unknown>:10: message6' in app.status.getvalue()

    node.source, node.line = (None, None)
    logger.info('message7', location=node)
    assert '\nmessage7' in app.status.getvalue()


@pytest.mark.sphinx('html', testroot='root')
def test_warning_location(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.warning('message1', location='index')
    assert 'index.txt: WARNING: message1' in app.warning.getvalue()

    logger.warning('message2', location=('index', 10))
    assert 'index.txt:10: WARNING: message2' in app.warning.getvalue()

    logger.warning('message3', location=None)
    assert colorize('red', 'WARNING: message3') in app.warning.getvalue()

    node = nodes.Element()
    node.source, node.line = ('index.txt', 10)
    logger.warning('message4', location=node)
    assert 'index.txt:10: WARNING: message4' in app.warning.getvalue()

    node.source, node.line = ('index.txt', None)
    logger.warning('message5', location=node)
    assert 'index.txt:: WARNING: message5' in app.warning.getvalue()

    node.source, node.line = (None, 10)
    logger.warning('message6', location=node)
    assert '<unknown>:10: WARNING: message6' in app.warning.getvalue()

    node.source, node.line = (None, None)
    logger.warning('message7', location=node)
    assert colorize('red', 'WARNING: message7') in app.warning.getvalue()


@pytest.mark.sphinx('html', testroot='root')
def test_suppress_logging(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.warning('message1')
    with logging.suppress_logging():
        logger.warning('message2')
        assert 'WARNING: message1' in app.warning.getvalue()
        assert 'WARNING: message2' not in app.warning.getvalue()

    assert 'WARNING: message1' in app.warning.getvalue()
    assert 'WARNING: message2' not in app.warning.getvalue()


@pytest.mark.sphinx('html', testroot='root')
def test_pending_warnings(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.warning('message1')
    with logging.pending_warnings():
        # not logged yet (buffered) in here
        logger.warning('message2')
        logger.warning('message3')
        assert 'WARNING: message1' in app.warning.getvalue()
        assert 'WARNING: message2' not in app.warning.getvalue()
        assert 'WARNING: message3' not in app.warning.getvalue()

    # actually logged as ordered
    warnings = strip_escape_sequences(app.warning.getvalue())
    assert 'WARNING: message2\nWARNING: message3' in warnings


@contextmanager
def force_colors():
    forcecolor = os.environ.get('FORCE_COLOR', None)

    try:
        os.environ['FORCE_COLOR'] = '1'
        yield
    finally:
        if forcecolor is None:
            os.environ.pop('FORCE_COLOR', None)
        else:
            os.environ['FORCE_COLOR'] = forcecolor


def test_log_no_ansi_colors(tmp_path):
    with force_colors():
        wfile = tmp_path / 'warnings.txt'
        srcdir = TESTS_ROOT / 'roots' / 'test-nitpicky-warnings'
        argv = list(map(str, ['-b', 'html', srcdir, tmp_path, '-n', '-w', wfile]))
        retcode = build_main(argv)
        assert retcode == 0

        content = wfile.read_text(encoding='utf8')
        assert '\x1b[91m' not in content


@pytest.mark.sphinx('html', testroot='root')
def test_colored_logs(app: SphinxTestApp) -> None:
    app.verbosity = 2
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    # default colors
    logger.debug('message1')
    logger.verbose('message2')
    logger.info('message3')
    logger.warning('message4')
    logger.critical('message5')
    logger.error('message6')

    assert colorize('darkgray', 'message1') in app.status.getvalue()
    assert 'message2\n' in app.status.getvalue()  # not colored
    assert 'message3\n' in app.status.getvalue()  # not colored
    assert colorize('red', 'WARNING: message4') in app.warning.getvalue()
    assert 'CRITICAL: message5\n' in app.warning.getvalue()  # not colored
    assert colorize('darkred', 'ERROR: message6') in app.warning.getvalue()

    # color specification
    logger.debug('message7', color='white')
    logger.info('message8', color='red')
    assert colorize('white', 'message7') in app.status.getvalue()
    assert colorize('red', 'message8') in app.status.getvalue()


@pytest.mark.xfail(
    os.name != 'posix',
    reason='Parallel mode does not work on Windows',
)
def test_logging_in_ParallelTasks(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    def child_process():
        logger.info('message1')
        logger.warning('message2', location='index')

    tasks = ParallelTasks(1)
    tasks.add_task(child_process)
    tasks.join()
    assert 'message1' in app.status.getvalue()
    assert 'index.txt: WARNING: message2' in app.warning.getvalue()


@pytest.mark.sphinx('html', testroot='root')
def test_output_with_unencodable_char(app):
    class StreamWriter(codecs.StreamWriter):
        def write(self, object):
            self.stream.write(object.encode('cp1252').decode('cp1252'))

    logging.setup(app, StreamWriter(app.status), app.warning)
    logger = logging.getLogger(__name__)

    # info with UnicodeEncodeError
    app.status.truncate(0)
    app.status.seek(0)
    logger.info('unicode \u206d...')
    assert app.status.getvalue() == 'unicode ?...\n'


@pytest.mark.sphinx('html', testroot='root')
def test_prefixed_warnings(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)

    logger.warning('message1')
    with prefixed_warnings('PREFIX:'):
        logger.warning('message2')
        with prefixed_warnings('Another PREFIX:'):
            logger.warning('message3')
        logger.warning('message4')
    logger.warning('message5')

    assert 'WARNING: message1' in app.warning.getvalue()
    assert 'WARNING: PREFIX: message2' in app.warning.getvalue()
    assert 'WARNING: Another PREFIX: message3' in app.warning.getvalue()
    assert 'WARNING: PREFIX: message4' in app.warning.getvalue()
    assert 'WARNING: message5' in app.warning.getvalue()


def test_get_node_location_abspath():
    # Ensure that node locations are reported as an absolute path,
    # even if the source attribute is a relative path.

    relative_filename = Path('relative', 'path.txt')
    absolute_filename = relative_filename.resolve()

    n = nodes.Element()
    n.source = str(relative_filename)

    location = logging.get_node_location(n)

    assert location == f'{absolute_filename}:'


@pytest.mark.sphinx('html', testroot='root', confoverrides={'show_warning_types': True})
def test_show_warning_types(app: SphinxTestApp) -> None:
    logging.setup(app, app.status, app.warning)
    logger = logging.getLogger(__name__)
    logger.warning('message2')
    logger.warning('message3', type='test')
    logger.warning('message4', type='test', subtype='logging')

    warnings = strip_escape_sequences(app.warning.getvalue()).splitlines()

    assert warnings == [
        'WARNING: message2',
        'WARNING: message3 [test]',
        'WARNING: message4 [test.logging]',
    ]



"""Test the build process with gettext builder with the test root."""

from __future__ import annotations

import gettext
import re
import subprocess
from contextlib import chdir
from pathlib import Path
from subprocess import CalledProcessError
from typing import TYPE_CHECKING

import pytest

from sphinx.builders.gettext import Catalog, MsgOrigin
from sphinx.testing.util import SphinxTestApp

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp

_MSGID_PATTERN = re.compile(r'msgid "((?:\n|.)*?)"\nmsgstr', re.MULTILINE)


def get_msgids(pot: str) -> list[str]:
    matches = _MSGID_PATTERN.findall(pot)
    return [m.replace('"\n"', '') for m in matches[1:]]


def test_Catalog_duplicated_message() -> None:
    catalog = Catalog()
    catalog.add('hello', MsgOrigin('/path/to/filename', 1))
    catalog.add('hello', MsgOrigin('/path/to/filename', 1))
    catalog.add('hello', MsgOrigin('/path/to/filename', 2))
    catalog.add('hello', MsgOrigin('/path/to/yetanother', 1))
    catalog.add('world', MsgOrigin('/path/to/filename', 1))

    assert len(list(catalog)) == 2

    msg1, msg2 = list(catalog)
    assert msg1.text == 'hello'
    assert msg1.locations == [
        ('/path/to/filename', 1),
        ('/path/to/filename', 2),
        ('/path/to/yetanother', 1),
    ]
    assert msg2.text == 'world'
    assert msg2.locations == [('/path/to/filename', 1)]


@pytest.mark.sphinx(
    'gettext',
    testroot='root',
    srcdir='root-gettext',
)
def test_build_gettext(app: SphinxTestApp) -> None:
    # Generic build; should fail only when the builder is horribly broken.
    app.build(force_all=True)

    # Do messages end up in the correct location?
    # top-level documents end up in a message catalog
    assert (app.outdir / 'extapi.pot').is_file()
    # directory items are grouped into sections
    assert (app.outdir / 'subdir.pot').is_file()

    # regression test for https://github.com/sphinx-doc/sphinx/issues/960
    catalog = (app.outdir / 'markup.pot').read_text(encoding='utf8')
    assert 'msgid "something, something else, something more"' in catalog


@pytest.mark.sphinx(
    'gettext',
    testroot='root',
    srcdir='root-gettext',
)
def test_msgfmt(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    (app.outdir / 'en' / 'LC_MESSAGES').mkdir(parents=True, exist_ok=True)
    with chdir(app.outdir):
        try:
            args = [
                'msginit',
                '--no-translator',
                '-i',
                'markup.pot',
                '--locale',
                'en_US',
            ]
            subprocess.run(args, capture_output=True, check=True)
        except OSError:
            pytest.skip()  # most likely msginit was not found
        except CalledProcessError as exc:
            print(exc.stdout)
            print(exc.stderr)
            msg = f'msginit exited with return code {exc.returncode}'
            raise AssertionError(msg) from exc

        assert (app.outdir / 'en_US.po').is_file(), 'msginit failed'
        try:
            args = [
                'msgfmt',
                'en_US.po',
                '-o',
                str(Path('en', 'LC_MESSAGES', 'test_root.mo')),
            ]
            subprocess.run(args, capture_output=True, check=True)
        except OSError:
            pytest.skip()  # most likely msgfmt was not found
        except CalledProcessError as exc:
            print(exc.stdout)
            print(exc.stderr)
            msg = f'msgfmt exited with return code {exc.returncode}'
            raise AssertionError(msg) from exc

        mo = app.outdir / 'en' / 'LC_MESSAGES' / 'test_root.mo'
        assert mo.is_file(), 'msgfmt failed'

    _ = gettext.translation('test_root', app.outdir, languages=['en']).gettext
    assert _('Testing various markup') == 'Testing various markup'


@pytest.mark.sphinx(
    'gettext',
    testroot='intl',
    srcdir='gettext',
    confoverrides={'gettext_compact': False},
)
def test_gettext_index_entries(app: SphinxTestApp) -> None:
    # regression test for https://github.com/sphinx-doc/sphinx/issues/976
    app.build(filenames=[app.srcdir / 'index_entries.txt'])

    pot = (app.outdir / 'index_entries.pot').read_text(encoding='utf8')
    msg_ids = get_msgids(pot)

    assert msg_ids == [
        'i18n with index entries',
        'index target section',
        'this is :index:`Newsletter` target paragraph.',
        'various index entries',
        "That's all.",
        'Mailing List',
        'Newsletter',
        'Recipients List',
        'First',
        'Second',
        'Third',
        'Entry',
        'See',
    ]


@pytest.mark.sphinx(
    'gettext',
    testroot='intl',
    srcdir='gettext',
    confoverrides={'gettext_compact': False, 'gettext_additional_targets': []},
)
def test_gettext_disable_index_entries(app: SphinxTestApp) -> None:
    # regression test for https://github.com/sphinx-doc/sphinx/issues/976
    app.env._pickled_doctree_cache.clear()  # clear cache
    app.build(filenames=[app.srcdir / 'index_entries.txt'])

    pot = (app.outdir / 'index_entries.pot').read_text(encoding='utf8')
    msg_ids = get_msgids(pot)

    assert msg_ids == [
        'i18n with index entries',
        'index target section',
        'this is :index:`Newsletter` target paragraph.',
        'various index entries',
        "That's all.",
    ]


@pytest.mark.sphinx(
    'gettext',
    testroot='intl',
    srcdir='gettext',
)
def test_gettext_template(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    assert (app.outdir / 'sphinx.pot').is_file()

    result = (app.outdir / 'sphinx.pot').read_text(encoding='utf8')
    assert 'Welcome' in result
    assert 'Sphinx %(version)s' in result


@pytest.mark.sphinx('gettext', testroot='gettext-template')
def test_gettext_template_msgid_order_in_sphinxpot(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert (app.outdir / 'sphinx.pot').is_file()

    result = (app.outdir / 'sphinx.pot').read_text(encoding='utf8')
    assert re.search(
        (
            'msgid "Template 1".*'
            'msgid "This is Template 1\\.".*'
            'msgid "Template 2".*'
            'msgid "This is Template 2\\.".*'
        ),
        result,
        flags=re.DOTALL,
    )


@pytest.mark.sphinx('gettext', testroot='gettext-custom-output-template')
def test_gettext_custom_output_template(app: SphinxTestApp) -> None:
    app.build(force_all=True)
    assert (app.outdir / 'index.pot').is_file()

    result = (app.outdir / 'index.pot').read_text(encoding='utf8')
    assert 'EVEN MORE DESCRIPTIVE TITLE' in result


@pytest.mark.sphinx(
    'gettext',
    testroot='root',
    srcdir='root-gettext',
    confoverrides={'gettext_compact': 'documentation'},
)
def test_build_single_pot(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    assert (app.outdir / 'documentation.pot').is_file()

    result = (app.outdir / 'documentation.pot').read_text(encoding='utf8')
    assert re.search(
        (
            'msgid "Todo".*'
            'msgid "Like footnotes.".*'
            'msgid "The minute.".*'
            'msgid "Generated section".*'
        ),
        result,
        flags=re.DOTALL,
    )


@pytest.mark.sphinx(
    'gettext',
    testroot='intl_substitution_definitions',
    srcdir='gettext-subst',
    confoverrides={'gettext_compact': False, 'gettext_additional_targets': ['image']},
)
def test_gettext_prolog_epilog_substitution(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    assert (app.outdir / 'prolog_epilog_substitution.pot').is_file()
    pot = (app.outdir / 'prolog_epilog_substitution.pot').read_text(encoding='utf8')
    msg_ids = get_msgids(pot)

    assert msg_ids == [
        'i18n with prologue and epilogue substitutions',
        'This is content that contains |subst_prolog_1|.',
        'Substituted image |subst_prolog_2| here.',
        'subst_prolog_2',
        '.. image:: /img.png',
        'This is content that contains |subst_epilog_1|.',
        'Substituted image |subst_epilog_2| here.',
        'subst_epilog_2',
        '.. image:: /i18n.png',
    ]


@pytest.mark.sphinx(
    'gettext',
    testroot='intl_substitution_definitions',
    srcdir='gettext-subst',
    confoverrides={'gettext_compact': False, 'gettext_additional_targets': ['image']},
)
def test_gettext_prolog_epilog_substitution_excluded(app: SphinxTestApp) -> None:
    # regression test for https://github.com/sphinx-doc/sphinx/issues/9428
    app.build(force_all=True)

    assert (app.outdir / 'prolog_epilog_substitution_excluded.pot').is_file()
    pot = (app.outdir / 'prolog_epilog_substitution_excluded.pot').read_text(
        encoding='utf8'
    )
    msg_ids = get_msgids(pot)

    assert msg_ids == [
        'i18n without prologue and epilogue substitutions',
        'This is content that does not include prologue and epilogue substitutions.',
    ]


@pytest.mark.sphinx(
    'gettext',
    testroot='root',
    srcdir='gettext',
    confoverrides={
        'gettext_compact': False,
        'gettext_additional_targets': ['literal-block', 'doctest-block'],
    },
)
def test_gettext_literalblock_additional(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    assert (app.outdir / 'literalblock.pot').is_file()
    pot = (app.outdir / 'literalblock.pot').read_text(encoding='utf8')
    msg_ids = get_msgids(pot)

    assert msg_ids == [
        'i18n with literal block',
        'Correct literal block::',
        'this is\\nliteral block',
        'Missing literal block::',
        "That's all.",
        'included raw.txt',
        '===\\nRaw\\n===\\n\\n.. raw:: html\\n\\n   <iframe src=\\"https://sphinx-doc.org\\"></iframe>\\n\\n',
        'code blocks',
        "def main\\n   'result'\\nend",
        '#include <stdlib.h>\\nint main(int argc, char** argv)\\n{\\n    return 0;\\n}',
        'example of C language',
        '#include <stdio.h>\\nint main(int argc, char** argv)\\n{\\n    return 0;\\n}',
        'literal-block\\nin list',
        'test_code_for_noqa()\\ncontinued()',
        'doctest blocks',
        '>>> import sys  # sys importing\\n>>> def main():  # define main '
        "function\\n...     sys.stdout.write('hello')  # call write method of "
        "stdout object\\n>>>\\n>>> if __name__ == '__main__':  # if run this py "
        'file as python script\\n...     main()  # call main',
    ]



"""Convert Sigstore attestations to PEP 740.

See https://github.com/trailofbits/pypi-attestations.
"""

# resolution fails without betterproto and protobuf-specs
# /// script
# requires-python = ">=3.11"
# dependencies = [
#     "pypi-attestations==0.0.21",
#     "betterproto==2.0.0b6",
# ]
# ///

from __future__ import annotations

import json
import sys
from base64 import b64decode
from pathlib import Path

from pypi_attestations import Attestation, Distribution
from sigstore.models import Bundle
from sigstore.verify.policy import Identity

ROOT = Path(__file__).resolve().parent.parent
DIST = ROOT / 'dist'
bundle_path = Path(sys.argv[1])
signer_identity = sys.argv[2]

for line in bundle_path.read_bytes().splitlines():
    dsse_envelope_payload = json.loads(line)['dsseEnvelope']['payload']
    subjects = json.loads(b64decode(dsse_envelope_payload))['subject']
    for subject in subjects:
        filename = subject['name']
        assert (DIST / filename).is_file()

        # Convert attestation from Sigstore to PEP 740
        print(f'Converting attestation for {filename}')
        sigstore_bundle = Bundle.from_json(line)
        attestation = Attestation.from_bundle(sigstore_bundle)
        attestation_path = DIST / f'{filename}.publish.attestation'
        attestation_path.write_text(attestation.model_dump_json())
        print(f'Attestation for {filename} written to {attestation_path}')
        print()

        # Validate attestation
        dist = Distribution.from_file(DIST / filename)
        attestation = Attestation.model_validate_json(attestation_path.read_bytes())
        identity = Identity(identity=signer_identity)
        attestation.verify(identity=identity, dist=dist)
        print(f'Verified {attestation_path}')
